#!/usr/bin/env python

from __future__ import division

import scipy.io as sio

from ..tools.tools import *


def cum2x(x, y, maxlag=0, nsamp=0, overlap=0, flag='biased'):
    """
  Cross-covariance
  Parameters:
      x,y    - data vectors/matrices with identical dimensions
               if x,y are matrices, rather than vectors, columns are
               assumed to correspond to independent realizations,
               overlap is set to 0, and samp_seg to the row dimension.
      maxlag - maximum lag to be computed    [default = 0]
    samp_seg - samples per segment  [default = data_length]
     overlap - percentage overlap of segments [default = 0]
               overlap is clipped to the allowed range of [0,99].
       flag  - 'biased', biased estimates are computed  [default]
              'unbiased', unbiased estimates are computed.

  Output:
       y_cum - estimated cross-covariance
               E x^*(n)y(n+m),   -maxlag <= m <= maxlag
  """

    (lx, nrecs) = x.shape
    if (lx, nrecs) != y.shape:
        raise ValueError('x,y should have identical dimensions')

    if lx == 1:
        lx = nrecs
        nrecs = 1

    if maxlag < 0: raise ValueError('maxlag must be non-negative')
    if nrecs > 1: nsamp = lx
    if nsamp <= 0 or nsamp > lx: nsamp = lx
    if nrecs > 1: overlap = 0
    overlap = max(0, min(overlap, 99))

    overlap = np.fix(overlap / 100 * nsamp)
    nadvance = nsamp - overlap
    if nrecs == 1:
        nrecs = np.fix((lx - overlap) / nadvance)

    nlags = 2 * maxlag + 1
    zlag = maxlag
    y_cum = np.zeros([nlags, 1])

    if flag == 'biased':
        scale = np.ones([nlags, 1]) / nsamp
    else:
        scale = make_arr((range(lx - maxlag, lx + 1), range(lx - 1, lx - maxlag - 1, -1)), axis=1).T
        scale = np.ones([2 * maxlag + 1, 1]) / scale

    ind = np.arange(nsamp).T
    for k in np.arange(nrecs):
        xs = x[ind].ravel(order='F')
        xs = xs - np.mean(xs)
        ys = y[ind].ravel(order='F')
        ys = ys - np.mean(ys)

        y_cum[zlag] = y_cum[zlag] + np.dot(xs, ys)

        for m in np.arange(1, maxlag + 1):
            y_cum[zlag - m] = y_cum[zlag - m] + np.dot(xs[m:nsamp].T, ys[0:nsamp - m])
            y_cum[zlag + m] = y_cum[zlag + m] + np.dot(xs[0:nsamp - m].T, ys[m:nsamp])

        ind = ind + int(nadvance)

    y_cum = y_cum * scale / nrecs

    return y_cum


def test():
    y = sio.loadmat(here(__file__) + '/demo/ma1.mat')['y']

    # The right results are:
    #           "biased": [--0.25719  -0.12011   0.35908   1.01378   0.35908  -0.12011  -0.25719]
    #           "unbiased": [-0.025190  -0.011753   0.035101   0.099002   0.035101  -0.011753  -0.025190]
    print(cum2x(y, y, 3, 100, 0, "biased"))
    print(cum2x(y, y, 3, 100, 0, "unbiased"))


if __name__ == '__main__':
    test()



#!/usr/bin/env python


import os

import numpy as np


def nextpow2(num):
    '''
  Returns the next highest power of 2 from the given value.
  Example
  -------
  >>nextpow2(1000)
  1024
  >>nextpow2(1024)
  2048

  Taken from: https://github.com/alaiacano/frfft/blob/master/frfft.py
  '''

    npow = 2
    while npow <= num:
        npow = npow * 2
    return npow


def flat_eq(x, y):
    """
  Emulate MATLAB's assignment of the form
  x(:) = y
  """
    z = x.reshape(1, -1)
    z = y
    return z.reshape(x.shape)


def make_arr(arrs, axis=0):
    """
  Create arrays like MATLAB does
  python                                 MATLAB
  make_arr((4, range(1,10)), axis=0) => [4; 1:9]
  """
    a = []
    ctr = 0
    for x in arrs:
        if len(np.shape(x)) == 0:
            a.append(np.array([[x]]))
        elif len(np.shape(x)) == 1:
            a.append(np.array([x]))
        else:
            a.append(x)
        ctr += 1
    return np.concatenate(a, axis)


def shape(o, n):
    """
  Behave like MATLAB's shape
  """
    s = o.shape
    if len(s) < n:
        x = tuple(np.ones(n - len(s)))
        return s + x
    else:
        return s


def here(f=__file__):
    """
  This script's directory
  """
    return os.path.dirname(os.path.realpath(f))



import mne.filter
import numpy as np
from scipy.signal import hilbert
from scipy.stats import iqr
from PyEMD import EMD
import antropy as ant
import pywt
from pyentrp import entropy as ent
from ..HOSA.conventional.bicoherence import bicoherence
from pyts.metrics.dtw import dtw
from scipy import signal
from scipy.fftpack import fft
from statsmodels.tsa.arima.model import ARIMA
from pactools.comodulogram import Comodulogram
#from pactools.utils import rayleigh_z
from EntropyHub import FuzzEn
from nilearn.connectome import ConnectivityMeasure
from .pdc_dtf import calculate_dtf_pdc
from fooof import FOOOFGroup
from scipy.stats import norm
from fooof.bands import Bands
from fooof.analysis import get_band_peak_fg
import matplotlib.pyplot as plt
from mne_features.univariate import compute_pow_freq_bands
from scipy.signal import welch
from mne_connectivity import spectral_connectivity_epochs, envelope_correlation
from scipy.signal import hilbert, coherence
from pybispectra import AAC, compute_tfr
# import tftb





def compute_DFA(data, sfreq=250, win_times=1):
    """
    Perform Detrended Fluctuation Analysis (DFA) to find long-term statistical correlations in a time series.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        sfreq (int, optional): Sampling frequency. Defaults to 250.
        win_times (int, optional): Window duration in seconds. Defaults to 1.

    Returns:
        ndarray: DFA features with shape (n_channels, section_num).

    References:
        Peng C K, Havlin S, Stanley H E, Goldberger A L. Quantification of scaling exponents and crossover phenomena in
        nonstationary heartbeat time series[J]. Chaos: An Interdisciplinary Journal of Nonlinear Science, 1995, 5(1): 82-87.
    """
    win_len = sfreq * win_times
    n_channel, n_times = data.shape
    section_num = n_times // win_len
    feature = np.zeros((n_channel, section_num))
    for i_section in range(section_num):
        for i_channel in range(n_channel):
            feature[i_channel,i_section] = ant.detrended_fluctuation(data[i_channel, i_section * win_len:(i_section+ 1) * win_len])
    feature = feature.T.reshape(-1)
    return feature
def compute_Shannon_entropy(data, sfreq=250,round_para=None, win_times=1):
    """
    Compute the Shannon entropy of the data.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        sfreq (int, optional): Sampling frequency. Defaults to 250.
        round_para (int, optional): Rounding precision for data. Defaults to None, default retention of all digits for calculation.
        win_times (int, optional): Window duration in seconds. Defaults to 1.

    Returns:
        ndarray: Shannon entropy features with shape (n_channels, section_num).

    References:
        Shannon C E. A mathematical theory of communication[J]. Bell System Technical Journal, 1948, 27(3): 379-423.
    """
    if round_para is not None:
        data = np.round(data, round_para)
    win_len = sfreq * win_times

    n_channel, n_times = data.shape
    section_num = n_times // win_len
    feature = np.zeros((n_channel, section_num))
    for i_section in range(section_num):
        for i_channel in range(n_channel):
            feature[i_channel,i_section]=ent.shannon_entropy(data[i_channel, i_section * win_len:(i_section+ 1) * win_len])
    feature = feature.T.reshape(-1)
    return feature
def Tsallis_Entropy(time_series,alpha=2):
    """
    Compute the Tsallis entropy of the sample data.

    Args:
        time_series (Union[Vector, str]): Input time series data.
        alpha (float, optional): Entropy parameter. Defaults to 2.

    Returns:
        float: Tsallis entropy value.

    References:
        Tsallis C. Possible generalization of Boltzmann-Gibbs statistics[J]. Journal of Statistical Physics, 1988, 52(1-2): 479-487.
        https://zhuanlan.zhihu.com/p/81462898 (Chinese reference explaining the base of the logarithm used)

    Note:
        There is a question about why the base of the logarithm used is 2.
    """
    # Check if string
    if not isinstance(time_series, str):
        time_series = list(time_series)
    # Create a frequency data
    data_set = list(set(time_series))
    freq_list = []
    for entry in data_set:
        counter = 0.
        for i in time_series:
            if i == entry:
                counter += 1
        freq_list.append(float(counter) / len(time_series))

    # Shannon entropy
    ent = 0.0
    for freq in freq_list:
        ent += freq ** (alpha)
    ent =1/(1-alpha)*(ent-1)
    return ent

def Renyi_Entropy(time_series,alpha):
    """
     Compute the Renyi entropy of the sample data.

     Args:
         time_series (Union[Vector, str]): Input time series data.
         alpha (float): Entropy parameter.

     Returns:
         float: Renyi entropy value.
     """
    # Check if string
    if not isinstance(time_series, str):
        time_series = list(time_series)

    # Create a frequency data
    data_set = list(set(time_series))
    freq_list = []
    for entry in data_set:
        counter = 0.
        for i in time_series:
            if i == entry:
                counter += 1
        freq_list.append(float(counter) / len(time_series))

    # Shannon entropy
    ent = 0.0
    for freq in freq_list:
        ent += freq**(alpha)
    ent=1/(1-alpha)*np.log2(ent)

    return ent
def compute_ARMA_kalman_filter(data,AR_p=10,MA_q=1):
    """
    Compute ARMA modeling coefficients using the Kalman filter.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        AR_p (int, optional): Order of the AR model. Defaults to 10.
        MA_q (int, optional): Order of the MA model. Defaults to 1.

    Returns:
        ndarray: ARMA coefficients with shape (n_channel, AR_p + MA_q).

    Example:
        >>> rng = np.random.RandomState(42)
        >>> n_epochs, n_channels, n_times = 2, 1, 1750
        >>> X = rng.randn(n_epochs, n_channels, n_times)
        >>> feat = Feature(X, sfreq=250, selected_funcs={'ARMA_kalman_filter'})
        >>> print(feat.features.shape)

    References:
        [1] Rossow A B, Salles E O T, Côco K F. Automatic sleep staging using a single-channel EEG modeling by Kalman filter
        and HMM[C]//ISSNIP Biosignals and Biorobotics Conference 2011. IEEE, 2011: 1-6.
    """
    n_channel, n_times = data.shape
    feature=np.zeros((n_channel,AR_p+MA_q))
    for i_channel in range(n_channel):
        arma_mod = ARIMA(data[i_channel,:], order=(AR_p, 0, MA_q))
        arma_res = arma_mod.fit()
        feature[i_channel,:]=np.concatenate([arma_res.polynomial_ar[1:], arma_res. polynomial_ma[1:]])
    feature = feature.T.reshape(-1)
    return feature

def get_fft_values(y, N=None, f_s=250):
    """
    Compute the FFT values of a time series.

    Args:
        y (array): Input time series data.
        N (int, optional): Number of samples to use. Defaults to None.
        f_s (int, optional): Sampling frequency. Defaults to 250.

    Returns:
        tuple: Frequencies and FFT values.

    """
    if N is None:
        N=y.shape[0]
    f_values = np.linspace(0.0, f_s/2.0, N//2)
    fft_values_ = fft(y)
    fft_values = 2.0/N * np.abs(fft_values_[0:N//2])
    return f_values, fft_values

def find_nearest(array, value):
    """
    Find the index of the nearest value in an array.

    Args:
        array (array): Input array.
        value (float): Value to find the nearest index for.

    Returns:
        int: Index of the nearest value in the array.
    """
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx

def compute_Harmonic_Parameters(data,sfreq=250,
                                band=np.array([[2, 3.8], [4, 7], [8, 13], [14, 30], [31, 48]])):
   """
    Compute harmonic parameters including center frequency, bandwidth, and spectral value at the center frequency.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        sfreq (int, optional): Sampling frequency. Defaults to 250.
        band (ndarray, optional): Frequency bands with shape (n_bands, 2). Defaults to predefined bands.

    Returns:
        ndarray: Harmonic parameters with shape (n_channels, n_band).
    """
   n_channel, n_times = data.shape
   band_num=band.shape[0]
   feature=np.zeros((n_channel,band_num))
   for i_channel in range(n_channel):
     f_, fft_ = get_fft_values(data[i_channel, :], f_s=sfreq)
     for i_band in range(band_num):
           center_frequency=(band[i_band,0] + band[i_band,1]) / 2
           frequency_band=np.abs(band[i_band,0] - band[i_band,1])
         #  feature[i_channel,i_band,0]=center_frequency
         #  feature[i_channel, i_band, 1] = frequency_band
           f_idx=find_nearest(f_,center_frequency)
           feature[i_channel, i_band]=fft_[f_idx]
   feature = feature.T.reshape(-1)
   return feature
def compute_Median_Frequency(data,sfreq=250,
                                band=np.array([[0.5,2],[2, 4], [4, 5],
                                [5, 7], [7, 10], [10, 13],[13,15],[15,20],[20,30],[30,40]]),N=None):
    """
    Compute the median frequency for each channel and frequency band.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        sfreq (int, optional): Sampling frequency. Defaults to 250.
        band (ndarray, optional): Frequency bands. Defaults to predefined bands.
        N (int, optional): Number of samples to use. Defaults to None.

    Returns:
        ndarray: Median frequency values with shape (n_channels, band_num).

    References:
        [1] Gudmundsson S, Runarsson T P, Sigurdsson S. Automatic sleep staging using support vector machines with posterior probability estimates
            [C]//International Conference on Computational Intelligence for Modelling, Control and Automation and International Conference on Intelligent Agents,
            Web Technologies and Internet Commerce (CIMCA-IAWTIC'06). IEEE, 2005, 2: 366-372.
        [2] Thongpanja S, Phinyomark A, Phukpattaranont P, et al. Mean and median frequency of EMG signal to determine muscle force based on
            time-dependent power spectrum[J]. Elektronika ir Elektrotechnika, 2013, 19(3): 51-56.
    """
    n_channel, n_times = data.shape
    band_num = band.shape[0]
    feature = np.zeros((n_channel, band_num))
    for i_channel in range(n_channel):
       f_, fft_ = get_fft_values(data[i_channel, :],N=N, f_s=sfreq)
       feature[i_channel,:]=band_Median_Frequency(Pxx=fft_,f=f_,band=band)
    feature = feature.T.reshape(-1)
    return feature
def band_Median_Frequency(Pxx, f, band=None):
    """
    Compute the median frequency for a given power spectral density (PSD) and frequency band.

    Args:
        Pxx (array): Power spectral density values.
        f (array): Frequency values.
        band (ndarray, optional): Selected frequency bands. Defaults to None.

    Returns:
        ndarray: Median frequency values for each band.
    """
    fea_num=int(band.shape[0])
    Median_Frequency=np.empty((fea_num));
    for i in range(fea_num):
        psd_m = 0
        idx = np.where((f >= band[i, 0]) & (f <= band[i, 1]))[0]
        psd= np.sum(Pxx[idx]**2)
        psd_half=psd/2
        Median_Frequency[i] = band[i, 0]

        for i_idx in idx:
            if(psd_m < psd_half):
                psd_m=psd_m+np.multiply(Pxx[i_idx], Pxx[i_idx])
            else:
                Median_Frequency[i]=f[i_idx]
                break

    return Median_Frequency

def filter_bank(data,sfreq=250,frequences=None):
    """
    Apply a filter bank to the input data.

    Args:
        data (ndarray): Input data with shape (n_channel, n_times).
        sfreq (int, optional): Sampling frequency. Defaults to 250.
        frequences (ndarray, optional): Frequency ranges for the filters. Shape (n_filters, 2). Defaults to None.

    Returns:
        ndarray: Filtered data with shape (n_filters, n_channel, n_times).
    """
    n_filters=frequences.shape[0]
    n_channel,n_times=data.shape
    filters_data=np.zeros((n_filters,n_channel,n_times))
    for i_filters in range(n_filters):
        b, a = signal.butter(8, [2*frequences[i_filters,0]/sfreq, 2*frequences[i_filters,1]/sfreq], 'bandpass')  # 配置滤波器 8 表示滤波器的阶数
        filters_data[i_filters, :] = signal.filtfilt(b, a, data)  # data为要过滤的信号
    return filters_data

def compute_Coherence(data,Co_channel=None,
            sfreq=250,band=np.array([[2, 3.8], [4, 7], [8, 13], [14, 30], [31, 48]])):
    """
    Compute coherence between different channels within specified frequency bands.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        Co_channel (ndarray, optional): Channels to compute coherence for. Shape (n_Co_channel, 2). Defaults to None.
        sfreq (int, optional): Sampling frequency. Defaults to 250.
        band (ndarray, optional): Frequency bands. Shape (fea_num, 2). Defaults to predefined bands.

    Returns:
        ndarray: Coherence features with shape (n_channel, n_channel * band_num).

    Notes:
        For single-channel data, this function is not applicable.
    """
    n_channel, n_times = data.shape
    band_num=band.shape[0]
    feature = np.zeros((n_channel, n_channel ,band_num))
    if Co_channel is None:
        Co_channel=np.zeros((n_channel*n_channel,2),dtype=np.int32)
        ij_channel=0
        for i_channel in range(n_channel):
            for j_channel in range(n_channel):
                Co_channel[ij_channel,:]=[i_channel,j_channel]
                ij_channel=ij_channel+1

    for i_Co_channel in range(Co_channel.shape[0]):
           channel_0=Co_channel[i_Co_channel,0];  channel_1=Co_channel[i_Co_channel,1];
           x=data[channel_0,:];     y=data[channel_1,:]
           print(x.shape,y.shape)
           ff,cxx=signal.coherence(x, y, fs=sfreq,
                            window='hann', nperseg=None, noverlap=None, nfft=None, detrend='constant', axis=- 1)
           feature[channel_0,channel_1,:]=band_DE(cxx,ff,Par_ratios=0,band=band)
    feature = feature.T.reshape(-1)
    return feature
# def  compute_WignerVilleDistribution(data,sfreq=250 ):
#     '''
#     reference:Şen B, Peker M, Çavuşoğlu A, Çelebi FV. A comparative study on classification of sleep stage based
#     on EEG signals using feature selection and classification algorithms.
#     J Med Syst. 2014 Mar;38(3):18. doi: 10.1007/s10916-014-0018-0. Epub 2014 Mar 9. PMID: 24609509.
#     :param data: ndarray, shape (n_channels, n_times)
#     :param sfreq:
#     :return:     ndarray, shape (n_channels, 4)
#     这里对于最大频率的理解存疑
#     '''
#     n_channel, n_times = data.shape
#     feature = np.zeros((n_channel, 4))
#     for i_channel in range(n_channel):
#         wvd=tftb.processing.WignerVilleDistribution(signal=data[i_channel,:],
#                                                 timestamps=np.arange(n_times)*(1/sfreq))
#         tfr_wvd, t_wvd, f_wvd = wvd.run()
#         feature[i_channel,:]=np.polyfit(t_wvd, tfr_wvd[-1,:], 3)
#     feature = feature.T.reshape(-1)
#     return feature

def compute_Renyi_Entropy(data, sfreq=250,round_para=None, win_times=1,alpha=2):
    """
     Compute the Renyi entropy for each channel using a sliding window approach.

     Args:
         data (ndarray): Input data with shape (n_channels, n_times).
         sfreq (int, optional): Sampling frequency. Defaults to 250.
         round_para (int, optional): Number of decimal places to round the data.  Defaults to None, default retention of all digits for calculation.
         win_times (int, optional): Window duration in seconds. Defaults to 1.
         alpha (float, optional): Renyi entropy parameter. Defaults to 2.

     Returns:
         ndarray: Computed Renyi entropy with shape (n_channels, section_num).

     Notes:
         - The entropy is calculated for each window of data and then averaged across all windows.
     """
    if round_para is not None:
        data = np.round(data, round_para)
    win_len = sfreq * win_times

    n_channel, n_times = data.shape
    section_num = n_times // win_len
    feature = np.zeros((n_channel, section_num))
    for i_section in range(section_num):
        for i_channel in range(n_channel):
            feature[i_channel,i_section]=Renyi_Entropy(data[i_channel, i_section * win_len:(i_section+ 1) * win_len],alpha=alpha)
    feature = feature.T.reshape(-1)
    return feature
def compute_Tsallis_Entropy(data, sfreq=250,round_para=None, win_times=1,alpha=2):
    """
    Compute the Tsallis entropy for each channel using a sliding window approach.

    Args:
      data (ndarray): Input data with shape (n_channels, n_times).
      sfreq (int, optional): Sampling frequency. Defaults to 250.
      round_para (int, optional): Number of decimal places to round the data.  Defaults to None, default retention of all digits for calculation.
      win_times (int, optional): Window duration in seconds. Defaults to 1.
      alpha (float, optional): Tsallis entropy parameter. Defaults to 2.

    Returns:
      ndarray: Computed Tsallis entropy with shape (n_channels, section_num).

    Notes:
      - The entropy is calculated for each window of data and then averaged across all windows.
    """
    if round_para is not None:
        data = np.round(data, round_para)
    win_len = sfreq * win_times

    n_channel, n_times = data.shape
    section_num = n_times // win_len
    feature = np.zeros((n_channel, section_num))
    for i_section in range(section_num):
        for i_channel in range(n_channel):
            feature[i_channel,i_section]=Tsallis_Entropy(data[i_channel, i_section * win_len:(i_section+ 1) * win_len],alpha=alpha)
    feature = feature.T.reshape(-1)
    return feature


def compute_Hilbert_abs(data):
    """
    Compute the absolute value of the Hilbert transform (envelope) of the data. (abandon)

    Args:
     data (ndarray): Input data with shape (n_channels, n_times).

    Returns:
     ndarray: Absolute value of the Hilbert transform with shape (n_channels,).

    Notes:
     - This function is currently deprecated.
    """
    n_channel, n_times = data.shape
    feature = abs(hilbert(data)).reshape(-1)
    # feature=feature[np.newaxis, :];
    return feature
def compute_EMD(data, sfreq=250, EMD_times=1, EMD_params=6):
    """
    Compute the Empirical Mode Decomposition (EMD) of the data.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        sfreq (int, optional): Sampling frequency. Defaults to 250.
        EMD_times (int, optional): Duration of each EMD segment in seconds. Defaults to 1.
        EMD_params (int, optional): Number of Intrinsic Mode Functions (IMFs) to extract. Defaults to 6.

    Returns:
        ndarray: EMD features with shape (n_channels, section_num * EMD_params * EMD_length).

    Notes:
        - The EMD is applied to segments of the data, and the resulting IMFs are used as features.
    """
    EMD_length = sfreq * EMD_times
    n_channel, n_times = data.shape
    section_num = n_times // EMD_length
    signal_imfs = np.zeros((n_channel, section_num, EMD_params, EMD_length))
    emd = EMD()
    for N_length in range(section_num):
        for N_channel in range(n_channel):
            IMFs = emd.emd(data[N_channel, N_length * EMD_length:(N_length + 1) * EMD_length])
            signal_imfs[N_channel, N_length, :, :] = IMFs[0:EMD_params, :]
    signal_imfs = signal_imfs.reshape((n_channel, -1))
    feature = signal_imfs.T.reshape(-1)
    return feature
def compute_hosa_bicoherence(data,nfft=None, wind=None, nsamp=None, overlap=None):
    """
    Compute the higher-order spectral analysis (HOSA) bicoherence of the data.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        nfft (int, optional): FFT length. Defaults to 128.
        wind (array-like, optional): Time-domain window to be applied. Defaults to None, which uses a Hanning window.
        nsamp (int, optional): Samples per segment. Defaults to None.
        overlap (float, optional): Percentage overlap. Defaults to 50.

    Returns:
        ndarray: Bicoherence features with shape (n_channels, nfft * nfft).

    Notes:
        - This function is experimental and may have issues.
    """
    n_channel,n_times=data.shape
    feature=[]
    for N_channel in range(n_channel):
        y = data[N_channel, :]
        if y.ndim == 1:
            y = y[np.newaxis, :]
        bic, _ = bicoherence(y, nfft, wind, nsamp, overlap)
        feature.append(bic)
    feature = np.array(feature).reshape((n_channel,-1))
    feature = feature.T.reshape(-1)
    return feature
def compute_Itakura_Distance(data,baseline_data = None,dist='square', options={'max_slope': 2.0},
                                    precomputed_cost=None, return_cost=False,
                                    return_accumulated=False, return_path=False):
    """
    Compute the Itakura distance between the data and baseline data using dynamic time warping (DTW).

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        baseline_data (ndarray, optional): Baseline data with shape (n_channels, n_times). Defaults to None.
        dist (str or callable, optional): Distance metric to use. Defaults to 'square'.
        options (dict, optional): Method options. Defaults to {'max_slope': 2.0}.
        precomputed_cost (array-like, optional): Precomputed cost matrix. Defaults to None.
        return_cost (bool, optional): If True, return the cost matrix. Defaults to False.
        return_accumulated (bool, optional): If True, return the accumulated cost matrix. Defaults to False.
        return_path (bool, optional): If True, return the optimal path. Defaults to False.

    Returns:
        ndarray: Itakura distance with shape (n_channels,).

    References:
        https://pyts.readthedocs.io/en/stable/generated/pyts.metrics.dtw.html#pyts.metrics.dtw
    """
    n_channel, n_times = data.shape
    Itakura_distance = np.zeros((n_channel))
    for i_channel in range(n_channel):
        Itakura_distance[i_channel] = dtw(x=data[i_channel, :], y=baseline_data[i_channel, :],
                                          dist=dist, method='itakura',
                                          options=options,
                                          precomputed_cost=precomputed_cost, return_cost=return_cost,
                                          return_accumulated=return_accumulated, return_path=return_path)
    return Itakura_distance




def compute_wavelet_entropy(data,sfreq=250,m_times=1,m_Par_ratios=1,m_entropy=True,Average=True,
                            wavelet_name= 'gaus1', band=np.array([[2, 3.8], [4, 7], [8, 13], [14, 30], [31, 48]])):
    """
    Computes wavelet entropy for given data.

    Args:
        data (ndarray): Input data of shape (n_channels, n_times).
        sfreq (float): Sampling frequency of the data.
        m_times (float): Time unit in seconds.
        m_Par_ratios (int): Whether to include ratios (1) or not (0).
        m_entropy (bool): Whether to use entropy formula (True) or energy formula (False).
        Average (bool): Whether to compute average or not.
        wavelet_name (str): Name of the wavelet to use.
        band (ndarray): Frequency bands to analyze, shape (fea_num, 2).

    Returns:
        ndarray: Computed features. If Average=True, shape is (n_channels, fea_num + m_Par_ratios * 2).
                 If Average=False, shape is (n_channels, (fea_num + m_Par_ratios * 2) * section_num).
    """
    time_sec = int(m_times * sfreq)
    fea_num = int(band.shape[0] + m_Par_ratios * 2)
    n_channel, n_times = data.shape
    section_num = int(np.ceil((n_times / time_sec)))
    if Average:
        de = np.empty((int(n_channel), fea_num))
    else:
        de = np.empty((int(n_channel), fea_num, section_num))
    for channel in range(n_channel):
        # initialization
        section_de = np.empty((fea_num, section_num))
        # for one second calculate cwt
        for section in range(section_num):
            section_data = data[channel, section * time_sec:(section + 1) * time_sec]
            spec, f = imp_extract_wavelet(section_data=section_data, Fs=sfreq, time_sec=time_sec, wavelet_name=wavelet_name)
            section_de[:, section] = band_DE(spec, f, Par_ratios=m_Par_ratios, band=band)
        section_de=Processing_inf_nan(section_de)
        if Average:
            de_mean = np.sum(section_de, axis=1)
        else:
            de_mean = section_de
        if m_entropy:
            de_mean[de_mean == 0] = 1e-6
            de_mean = np.multiply(de_mean, np.log(de_mean))
        de[channel, :] = de_mean
    feature = de.T.reshape(-1)
    return feature
def imp_extract_wavelet(section_data,Fs, time_sec,wavelet_name):
    """
    Extracts wavelet transform from the data.

    Args:
        section_data (ndarray): Data segment to analyze.
        Fs (float): Sampling frequency.
        time_sec (int): Length of the time segment.
        wavelet_name (str): Name of the wavelet to use.

    Returns:
        tuple: (cwt_re, f1), where cwt_re is the wavelet transform, and f1 is the frequency vector.
    """
    f = np.arange(1, 129, 0.2)
    [wt, f1] = pywt.cwt(section_data, f, wavelet_name, 1 / Fs)  # 'mexh'
    cwt_re = np.sum(abs(wt), axis=1) * 2 / time_sec;  #
    return cwt_re,f1
def imp_extract_fft(section_data,Fs,time_sec):
    """
    Extracts FFT from the data.

    Args:
        section_data (ndarray): Data segment to analyze.
        Fs (float): Sampling frequency.
        time_sec (int): Length of the time segment.

    Returns:
        tuple: (m_fft, f), where m_fft is the FFT of the data, and f is the frequency vector.
    """
    f = np.arange(time_sec) / (time_sec / Fs)
    m_fft = abs(fft(section_data, time_sec) * 2 / time_sec);
    return m_fft[range(int(time_sec / 2))], f[range(int(time_sec/ 2))]
def band_DE(Pxx, f, Par_ratios=1, band=None):
    """
    Computes features from fixed frequency bands.

    Args:
        Pxx (ndarray): Power spectral density.
        f (ndarray): Frequency vector.
        Par_ratios (int): Whether to compute ratios (1) or not (0).
        band (ndarray): Frequency bands to analyze, shape (fea_num, 2).

    Returns:
        ndarray: Computed features.
    """
    fea_num=int(band.shape[0])
    psd = np.empty((fea_num))
    for i in range(fea_num):
        idx = np.where((f >= band[i, 0]) & (f <= band[i, 1]))
        psd[i] = np.sum(np.multiply(Pxx[idx], Pxx[idx]))
    if Par_ratios == 1:
        if psd[1]==0:
            psd[1]=1e-6
        san_D = np.hstack((psd, psd[2] / psd[1], psd[3] / psd[1]))
    else:
        san_D = psd
    return san_D

def Processing_inf_nan(data):
    """
    Processes infinite and NaN values in the data.

    Args:
        data (ndarray): Input data to process.

    Returns:
        ndarray: Processed data with infinite and NaN values handled.
    """
    data_inf = np.isinf(data)
    data[data_inf] = 0
    data_nan = np.isnan(data)
    data[data_nan] = np.max(data)
    return data

def compute_test2(data):
    """
    Computes the mean of the data along the last axis.

    Args:
        data (ndarray): Input data of shape (n_channels, n_times).

    Returns:
        ndarray: Mean values, shape (n_channels,).
    """
    return np.mean(data, axis=-1)



def compute_Num_zero_crossings(data):
    """
    Computes the number of zero crossings for each channel.

    Args:
        data (ndarray): Input data of shape (n_channels, n_times).

    Returns:
        ndarray: Number of zero crossings per channel, shape (n_channels,).
    """
    return ant.num_zerocross(data, axis=1)

def compute_Petrosian_fd(data):
    """
    Computes the Petrosian fractal dimension for each channel.

    Args:
        data (ndarray): Input data of shape (n_channels, n_times).

    Returns:
        ndarray: Fractal dimension per channel, shape (n_channels,).
    """
    return ant.petrosian_fd(data, axis=1)


def compute_perm_entropy(data):
    """
    Computes permutation entropy for each channel.

    Args:
        data (ndarray): Input data of shape (n_channels, n_times).

    Returns:
        ndarray: Permutation entropy per channel, shape (n_channels,).
    """
    return np.array([ant.perm_entropy(each_channel, normalize=True) for each_channel in data])


def compute_detrended_fluctuation(data):
    """
    Computes detrended fluctuation analysis for each channel.

    Args:
      data (ndarray): Input data of shape (n_channels, n_times).

    Returns:
      ndarray: Detrended fluctuation per channel, shape (n_channels,).
    """
    return np.array([ant.detrended_fluctuation(each_channel) for each_channel in data])


def compute_multiscale_sample_entropy(data, sample_length=1, tolerance=None, maxscale=None):
    """
    Computes multiscale sample entropy for each channel.

    Args:
        data (ndarray): Input data of shape (n_channels, n_times).
        sample_length (int): Length of the sample.
        tolerance (float): Tolerance level for sample entropy.
        maxscale (int): Maximum scale for multiscale entropy.

    Returns:
        ndarray: Multiscale sample entropy features, shape (n_channels, maxscale).
    """
    n_channel, n_times = data.shape
    if maxscale is None:
        maxscale = n_times
    multi_en_value = np.zeros((n_channel,maxscale))
    for i_channel in range(n_channel):
        multi_en_value[i_channel, :] = np.array(
            ent.multiscale_entropy(data[i_channel, :], sample_length=sample_length, tolerance=tolerance, maxscale=maxscale))
    feature = multi_en_value.T.reshape(-1)
    return feature


def compute_multiscale_permutation_entropy(data, m=1, delay=1, scale=1):
    """
    Compute the multiscale permutation entropy for each channel in the data.

    Args:
        data (ndarray): The input data array with shape (n_channels, n_times).
        m (int): Embedding dimension for permutation entropy. Default is 1.
        delay (int): Time delay for permutation entropy. Default is 1.
        scale (int): Scale factor for multiscale permutation entropy. Default is 1.

    Returns:
        ndarray: Multiscale permutation entropy features with shape (n_channels,).
    """
    n_channel, n_times = data.shape
    multi_per_en_value = np.zeros((n_channel,scale))
    for i_channel in  range(n_channel):
        multi_per_en_value[i_channel,:]= np.array(ent.multiscale_permutation_entropy(data[i_channel,:], m, delay, scale))
    feature = multi_per_en_value.T.reshape(-1)
    return feature


def compute_fuzzy_entropy(data,m=2, tau=1, r=(.2,2), Fx='default', Logx=np.exp(1)):
    """
    Compute fuzzy entropy for each channel in the data.

    Args:
        data (ndarray): The input data array with shape (n_channels, n_times).
        m (int): Embedding dimension. Default is 2.
        tau (int): Time delay. Default is 1.
        r (float or tuple): Fuzzy function parameters. Default is (.2, 2).
        Fx (str): Fuzzy function name. Default is 'default'.
        Logx (float): Base of the logarithm for normalization. Default is e.

    Returns:
        ndarray: Fuzzy entropy features with shape (n_channels,).
    """
    n_channel, n_times = data.shape
    FuzzEn_value = np.zeros((n_channel,m))
    for i_channel in  range(n_channel):
        FuzzEn_value[i_channel],_,_=FuzzEn(data[i_channel],m=m,tau=tau,Fx=Fx,r=r,Logx=Logx)
    feature = FuzzEn_value.T.reshape(-1)
    return feature





def compute_cross_frequency_coupling(data,sfreq=250,band=np.array([[1,4], [4,8],[8,10], [10,13], [13,20], [20,30], [30,45]]),
                 mode='eeg_rhythm', low_fq_range=None, low_fq_width=2., high_fq_range='auto',
                 high_fq_width='auto', method='tort', n_surrogates=0,n_jobs=1):
    """
        Compute cross-frequency coupling using either 'eeg_rhythm' or 'Fixed_width' mode.

        Args:
            data (ndarray): The input data array with shape (n_channels, n_times).
            sfreq (int): Sampling frequency of the time signal. Default is 250 Hz.
            band (ndarray): Frequency bands for analysis with shape (fea_num, 2). Default is predefined bands.
            mode (str): Mode of computation, either 'eeg_rhythm' or 'Fixed_width'. Default is 'eeg_rhythm'.
            low_fq_range (array or list): Filtering frequencies for phase signal in 'Fixed_width' mode. Default is None.
            low_fq_width (float): Bandwidth of the band-pass filter for phase signal. Default is 2.0.
            high_fq_range (array or list or 'auto'): Filtering frequencies for amplitude signal. Default is 'auto'.
            high_fq_width (float or 'auto'): Bandwidth of the band-pass filter for amplitude signal. Default is 'auto'.
            method (str): Method for computing modulation index. Default is 'tort'.
            n_surrogates (int): Number of surrogates for z-score computation. Default is 0.
            n_jobs (int): Number of parallel jobs. Default is 1.

        Returns:
            ndarray: Cross-frequency coupling features with shape (n_channels, band_num, band_num) or (n_channels, low_fq_range.shape[0], high_fq_range.shape[0]).

        Notes:
        - This function has been abandoned.
        """
    n_channel, n_times = data.shape
    if mode=='eeg_rhythm':
        band_num = band.shape[0]
        feature = np.zeros((n_channel, band_num, band_num))
        for i_band in range(band_num):
            i_center_fq = (band[i_band, 0] + band[i_band, 1]) / 2
            i_fq_width = band[i_band, 1] - band[i_band, 0]
            for j_band in range(band_num):
                    j_center_fq = (band[j_band, 0] + band[j_band, 1]) / 2
                    j_fq_width = band[j_band, 1] - band[j_band, 0]
                    c = Comodulogram(fs=sfreq, low_fq_range=np.array(i_center_fq), low_fq_width=i_fq_width,
                                     high_fq_range=np.array(j_center_fq),high_fq_width=j_fq_width,method=method, n_surrogates=n_surrogates,n_jobs=n_jobs)
                    for N_channel in range(n_channel):
                      sig=data[N_channel,:]
                      feature[N_channel,i_band,j_band]=c.fit(sig).comod_
        feature = feature.reshape(-1)
    if mode=='Fixed_width':
        feature_=np.zeros((n_channel,low_fq_range.shape[0],high_fq_range.shape[0]))
        c = Comodulogram(fs=sfreq, low_fq_range=low_fq_range, low_fq_width=low_fq_width,
                         high_fq_range=high_fq_range, high_fq_width=high_fq_width,method=method, n_surrogates=n_surrogates,n_jobs=n_jobs)
        for N_channel in range(n_channel):
            sig = data[N_channel, :]
            feature_[N_channel,:,:]=c.fit(sig).comod_
        feature=feature_.T.reshape(-1)
    return feature


# def  compute_stft_2019(data,sfreq=250,win_times=10,n_fre_idx=36):
#     """
#     Compute the short-time Fourier transform (STFT) and sum the energy at fixed frequency points.
#
#     Args:
#         data (ndarray): The input data array with shape (n_channels, n_times).
#         sfreq (int): Sampling frequency of the time signal. Default is 250 Hz.
#         win_times (int): Window length in seconds. Default is 10.
#         n_fre_idx (int): Number of frequency points to sum. Default is 36.
#
#     Returns:
#         ndarray: STFT-based features with shape (n_channels, section_num, n_fre_idx).
#     """
#     from scipy.signal import stft
#     win_len = sfreq * win_times
#     n_channel, n_times = data.shape
#     section_num = n_times // win_len
#     feature = np.zeros((n_channel, section_num,n_fre_idx))
#     for i_section in range(section_num):
#         for i_channel in range(n_channel):
#             X= data[i_channel, i_section * win_len:(i_section + 1) * win_len]
#             f, t, Zxx = stft(X, fs=sfreq, window='blackman', nperseg=256, noverlap=None, nfft=512,
#                              detrend=False, boundary='zeros', padded=True)
#             Y=10*np.log(abs(np.mean(Zxx[2:n_fre_idx*2+1:2,:],axis=1)))
#             feature[i_channel, i_section] =Y
#     feature = feature.T.reshape(-1)
#     return feature


def flatten_lower_triangle(matrix):
    """
    Flatten the lower triangle of a square matrix into a 1D array.

    Args:
        matrix (ndarray): Square matrix to flatten.

    Returns:
        ndarray: Flattened array with shape (n_channel*(n_channel-1)//2,).
    """
    rows = len(matrix)
    flattened = []
    for i in range(rows):
        for j in range(i):
            flattened.append(matrix[i][j])
    flattened = np.array(flattened)
    return flattened

def reshape_to_lower_triangle(flattened_array,n_channel):
    """
    Reshape a 1D array into the lower triangle of a square matrix.

    Args:
        flattened_array (ndarray): 1D array to reshape.
        n_channel (int): Number of channels, defining the size of the square matrix.

    Returns:
        ndarray: Square matrix with shape (n_channel, n_channel).
    """
    matrix=np.zeros((n_channel,n_channel))
    count = 0
    for i in range(n_channel):
        for j in range(i):
            matrix[i][j] = flattened_array[count]
            count += 1
    return matrix




def compute_correlation_matrix(data,sfreq=250,
                               kind="correlation",filter_bank=None,
                               n_win=1,log = False,standardize=True):
    """
    Compute various types of connectivity measures from EEG data.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        sfreq (int): Sampling frequency of the time signal. Default is 250 Hz.
        kind (str): Type of connectivity measure to compute. The available options are:
            - **Nilearn Measures**:
              - `"covariance"`: Measures the covariance between signals.
              - `"correlation"`: Measures the Pearson correlation coefficient between signals.
              - `"partial correlation"`: Measures the partial correlation, accounting for the influence of other signals.
              - `"tangent"`: Computes the tangent connectivity measure.  For the use of “tangent” see Varoquaux et al [1].
              - `"precision"`: Measures the precision of the connectivity.
           - **MNE-connectivity Measures**:
              - `"coh"`: Coherence.
              - `"cohy"`: Coherency.
              - `"imcoh"`: Imaginary part of Coherency.
              - `"cacoh"`: Canonical Coherency (CaCoh).
              - `"mic"`: Maximised Imaginary part of Coherency (MIC).
              - `"mim"`: Multivariate Interaction Measure (MIM).
              - `"plv"`: Phase-Locking Value (PLV).
              - `"ciplv"`: Corrected Imaginary PLV (ciPLV).
              - `"ppc"`: Pairwise Phase Consistency (PPC).
              - `"pli"`: Phase Lag Index (PLI).
              - `"pli2_unbiased"`: Unbiased estimator of squared PLI.
              - `"dpli"`: Directed PLI (DPLI).
              - `"wpli"`: Weighted PLI (WPLI).
              - `"wpli2_debiased"`: Debiased estimator of squared WPLI.
              - `"gc"`: State-space Granger Causality (GC).
              - `"gc_tr"`: State-space GC on time-reversed signals.
              - `"pec"`: power envolope correlation
           - **My-connectivity Measures**:
              - `"mcorrelation"`: Measures the Pearson correlation coefficient between signals.
              - `"mcoh"`: Coherence.
              - `"mplv"`: Phase-Locking Value (PLV).
        filter_bank (ndarray or list, optional): Band-pass filter parameters with shape (2,) [low_freq, high_freq]. Default is None (no filtering).
        n_win (int): Number of windows to split the data into. If the connectivity measure requires multiple epochs, this parameter helps in splitting one epoch into multiple parts. Default is 1.
        log (default False): If True , square and take the log before orthonalizing envelopes or computing correlations.
    Returns:
        ndarray: Flattened array of the computed connectivity matrix with shape (n_channel * n_channel,).

    Notes:
        - For certain measures like "tangent","plv", multiple epochs are required. Ensure `n_win` is set appropriately for such measures.
        - If the `filter_bank` is specified, the data is band-pass filtered before computing the connectivity.
        - In case of an error during connectivity computation, the function returns an identity matrix and prints a warning message. Ensure the parameters are set correctly to avoid computation errors.
    References:
        [1]Gael Varoquaux, Flore Baronnet, Andreas Kleinschmidt, Pierre Fillard, and Bertrand Thirion. Detection of brain functional-connectivity difference in post-stroke patients using group-level covariance modeling. In Tianzi Jiang, Nassir Navab, Josien P. W. Pluim, and Max A. Viergever, editors, Medical image computing and computer-assisted intervention - MICCAI 2010, Lecture notes in computer science, 200–208. Berlin, Heidelberg, 2010. Springer. https://link.springer.com/chapter/10.1007/978-3-642-15705-9_25.
    """

    n_channel,n_times=data.shape
    adjusted_n_times = (n_times // n_win) * n_win
    data = data[:,:adjusted_n_times]
    #tangent 这个是多个epoch放在一起才能计算的
    if kind in ["mcorrelation","mcoh","mplv"]:
        if filter_bank is not None:
            data = mne.filter.filter_data(data, sfreq=sfreq, l_freq=filter_bank[0], h_freq=filter_bank[1])
        feature = calculate_temp_correlation(data, sfreq, kind)
    elif kind in ["covariance","correlation", "partial correlation", "tangent","precision"]:
        if filter_bank is not None:
            data = mne.filter.filter_data(data, sfreq=sfreq, l_freq=filter_bank[0], h_freq=filter_bank[1])
        time_series = data.transpose(1, 0)
        time_series = time_series.reshape(n_win, n_times//n_win,n_channel)
        connectivity_measure = ConnectivityMeasure(kind=kind,standardize=standardize)
        matrix_0 = connectivity_measure.fit_transform(time_series)[0]
        feature = matrix_0
    elif kind in ['ciplv', 'ppc', 'pli', 'dpli', 'wpli', 'wpli2_debiased', 'cohy', 'imcoh','coh','plv','gc','gc_tr','mic','mim']:
        new_data = data.reshape([n_win,n_channel,n_times//n_win])
        try:
            if filter_bank is None:
                feature_1 = spectral_connectivity_epochs(data=new_data, method=kind, mode='multitaper', sfreq=sfreq,
                                                         faverage=True, mt_adaptive=False,verbose=False)
            else:
                feature_1=spectral_connectivity_epochs(data=new_data,method=kind,mode='multitaper', sfreq=sfreq, fmin=filter_bank[0],
                                                       fmax=filter_bank[1],faverage=True, mt_adaptive=False,verbose=False)
            feature_2 = np.squeeze(feature_1.get_data("dense"))
            np.fill_diagonal(feature_2,1)
            feature = feature_2 + feature_2.T - np.diag(feature_2.diagonal())
        except:
               feature = np.eye(n_channel)
               print("feature connectivity jump")
    elif kind in ['pec']:
        new_data = data.reshape([n_win, n_channel, n_times // n_win])
        feature_1 = envelope_correlation(data=new_data,log=log,verbose =False)
        feature = np.squeeze(feature_1.get_data("dense"))

    elif kind in ['dtf','dpc']:
        feature=calculate_dtf_pdc(data,sfreq=sfreq,kind=kind,p=None,normalize_=True,filter_bank=filter_bank)
    feature = feature.T.reshape(-1)
    return feature


def calculate_temp_correlation(data, sfreq=250, method="correlation"):
    """
    Calculte non-spectral connectivity "correlation","plv", "coh" measures from EEG data.
    Args:
        data: (ndarray): Input data with shape (n_channels, n_times).
        sfreq (int): Sampling frequency of the time signal. Default is 250 Hz.
        kind (str): Type of connectivity measure to compute. The available options are:"correlation","plv", "coh"

    Returns:
        correlation_matrix: (ndarray) (n_channels,n_channels)

    """
    num_channels = data.shape[0]
    correlation_matrix = np.zeros((num_channels, num_channels))
    for i in range(num_channels):
        for j in range(num_channels):
            if i != j:
                channel_data1 = data[i, :]
                channel_data2 = data[j, :]
                if method == "mcorrelation":
                    correlation_matrix[i, j] = calculate_channel_correlation_pearson(channel_data1, channel_data2)
                elif method == "mplv":
                    correlation_matrix[i, j] = calculate_channel_correlation_plv(channel_data1, channel_data2)
                elif method == "mcoh":
                    correlation_matrix[i, j] = calculate_channel_correlation_coh(channel_data1, channel_data2,sfreq)
            else:
                correlation_matrix[i, j] = 1  # 自相关为1
    return correlation_matrix

def compute_pac_connectivity(data,sfreq=250, method='tort', band=np.array([[4, 8],[30,45]]),
                             n_surrogates=0, mode="self", approach_pac="mean"):
    """
    Compute Phase-Amplitude Coupling (PAC) connectivity from EEG data.

    Args:
       data (ndarray): Input data with shape (n_channels, n_times).
       sfreq (int): Sampling frequency of the time signal. Default is 250 Hz.
       method (str): Method for computing PAC. Options are:
           - "tort": Tortoise method
           - "jiang": Jiang method
       band (ndarray): Frequency bands for PAC computation with shape (2, 2). Each row specifies the low and high frequencies for the band.
       n_surrogates (int): Number of surrogates for significance testing. Default is 0 (no surrogates).
       mode (str): Mode for PAC computation. Options are:
           - "self": Compute PAC for each channel with itself.
           - "non-self": Compute PAC between each pair of channels.
       approach_pac (str): Approach for summarizing PAC values. Options are:
           - "mean": Use the mean PAC value.
           - "max": Use the maximum PAC value.

    Returns:
       ndarray: Flattened array of PAC connectivity features. The shape depends on the `mode`:
           - If `mode` is "self": (n_channels,)
           - If `mode` is "non-self": (n_channels * n_channels,)

    Notes:
       - The `band` parameter specifies the frequency range for the low and high frequency bands used in PAC computation.
       - The `method` parameter determines the algorithm used for PAC calculation.
       - In "self" mode, PAC is computed for each channel individually.
       - In "non-self" mode, PAC is computed for every pair of channels.
       - The `approach_pac` parameter determines how the PAC values are aggregated: either by taking the mean or the maximum value.

    Example:
       To compute PAC using the "tort" method for each channel with itself, averaging the PAC values:
       ```python
       pac_features = compute_pac_connectivity(data, method='tort', mode='self', approach_pac='mean')
       ```
    """

    n_channel, n_times = data.shape
    low_fq_width = band[0, 1] - band[0, 0]
    high_fq_width = band[1, 1] - band[1, 0]
    low_fq_range =np.linspace(band[0,0],band[0,1],low_fq_width)
    high_fq_range = np.linspace(band[1, 0], band[1, 1], high_fq_width)
    c = Comodulogram(fs=sfreq, low_fq_range=low_fq_range, high_fq_range=high_fq_range,method=method,
                     n_surrogates=n_surrogates,n_jobs=-1)
    if mode== "self":
        feature=np.zeros((n_channel))
        for N_channel in range(n_channel):
            sig = data[N_channel, :]
            pac_matrix=c.fit(sig).comod_
            if approach_pac=="mean":
                  feature[N_channel] = np.mean(pac_matrix)
            elif approach_pac=="max":
                feature[N_channel] = np.max(pac_matrix)

    elif mode== "non-self":
        feature = np.zeros((n_channel,n_channel))
        for i_channel in range(n_channel):
            for j_channel in range(n_channel):
                sig_low = data[i_channel, :]
                sig_high = data[j_channel, :]
                pac_matrix = c.fit(sig_low,sig_high).comod_
                if approach_pac == "mean":
                    feature[i_channel,j_channel] =  np.mean(pac_matrix)
                elif approach_pac == "max":
                    feature[i_channel, j_channel] = np.max(pac_matrix)
    feature = feature.reshape(-1)
    return feature
def compute_aac_connectivity(data, sfreq=250, band=np.array([[4, 8],[30,45]]),tfr_mode="morlet", n_cycles=7,mode="self",approach_aac="mean",n_jobs=1):
    """
    Compute Amplitude-Amplitude Coupling (AAC) connectivity from EEG data.

    Args:
       data (ndarray): Input data with shape (n_channels, n_times).
       sfreq (int): Sampling frequency of the time signal. Default is 250 Hz.
       band (ndarray): Frequency bands for PAC computation with shape (2, 2). Each row specifies the low and high frequencies for the band.
       tfr_mode(str): mode of time frequncy representation Options are:
           - "morlet"      See mne.time_frequency.tfr_array_morlet()
           - "multitaper"  See mne.time_frequency.tfr_array_multitaper()
       mode (str): Mode for AAC computation. Options are:
           - "self": Compute AAC for each channel with itself.
           - "non-self": Compute AAC between each pair of channels.
       n_cycles(int|float): Number of cycles in the wavelet when computing the TFR. If an array, the number of cycles is given for each frequency, otherwise a fixed value across all frequencies is used.
       approach_aac (str): Approach for summarizing AAC values. Options are:
           - "mean": Use the mean AAC value.
           - "max": Use the maximum AAC value.

    Returns:
       ndarray: Flattened array of AAC connectivity features. The shape depends on the `mode`:
           - If `mode` is "self": (n_channels,)
           - If `mode` is "non-self": (n_channels * n_channels,)

    Notes:
       - The `band` parameter specifies the frequency range for the low and high frequency bands used in PAC computation.
       - In "self" mode, AAC is computed for each channel individually.
       - In "non-self" mode, PAC is computed for every pair of channels.
       - The `approach_aac` parameter determines how the AAC values are aggregated: either by taking the mean or the maximum value.
       -[i,j] i for (band[0,0]  band [0,1]) seed,  j for (band[1,0]  band [1,1]) target
    """
    n_channel, n_times = data.shape
    freqs = np.arange(max(band[0,0]-1,0), band[1,1]+1, 0.5)
    fft_coeffs, freqs = compute_tfr(data= np.expand_dims(data, axis=0), sampling_freq=sfreq, freqs=freqs,
                                    tfr_mode=tfr_mode ,n_cycles=n_cycles, verbose=False)
    aac = AAC(data=fft_coeffs, freqs=freqs, sampling_freq=sfreq, verbose=False)
    if mode == "self":
        ch_idx = tuple(range(n_channel))
        indices_self = (ch_idx, ch_idx)
        aac.compute(indices=indices_self,f1s=(band[0, 0], band[0, 1]),f2s=(band[1, 0], band[1, 1]),n_jobs=n_jobs)
        aac_results = aac.results.get_results(copy=False)
        if approach_aac == "mean":
            feature= np.mean(aac_results, axis=(1,2))
        elif approach_aac == "max":
            feature = np.max(aac_results, axis=(1,2))
        feature = feature.reshape(-1)
    elif mode== "non-self":
        aac.compute(f1s=(band[0, 0], band[0, 1]), f2s=(band[1, 0], band[1, 1]), n_jobs=n_jobs)
        aac_results = aac.results.get_results(copy=False)
        if approach_aac == "mean":
            feature=  np.mean(aac_results,axis=(1,2))
        elif approach_aac == "max":
            feature= np.max(aac_results,axis=(1,2))
        feature.reshape(n_channel,n_channel)
        feature = feature.T.reshape(-1)
    return feature
def compute_pac_connectivity_mod(data,sfreq=250, method='tort', band=np.array([[4, 8],[30,45]]),
                             n_surrogates=0, mode="self", approach_pac="mean"):
    """
    Compute Phase-Amplitude Coupling (PAC) connectivity from EEG data.

    Args:
       data (ndarray): Input data with shape (n_channels, n_times).
       sfreq (int): Sampling frequency of the time signal. Default is 250 Hz.
       method (str): Method for computing PAC. Options are:
           - "tort": Tortoise method
           - "jiang": Jiang method
       band (ndarray): Frequency bands for PAC computation with shape (2, 2). Each row specifies the low and high frequencies for the band.
       n_surrogates (int): Number of surrogates for significance testing. Default is 0 (no surrogates).
       mode (str): Mode for PAC computation. Options are:
           - "self": Compute PAC for each channel with itself.
           - "non-self": Compute PAC between each pair of channels.
       approach_pac (str): Approach for summarizing PAC values. Options are:
           - "mean": Use the mean PAC value.
           - "max": Use the maximum PAC value.

    Returns:
       ndarray: Flattened array of PAC connectivity features. The shape depends on the `mode`:
           - If `mode` is "self": (n_channels,)
           - If `mode` is "non-self": (n_channels * n_channels,)

    Notes:
       - The `band` parameter specifies the frequency range for the low and high frequency bands used in PAC computation.
       - The `method` parameter determines the algorithm used for PAC calculation.
       - In "self" mode, PAC is computed for each channel individually.
       - In "non-self" mode, PAC is computed for every pair of channels.
       - The `approach_pac` parameter determines how the PAC values are aggregated: either by taking the mean or the maximum value.

    Example:
       To compute PAC using the "tort" method for each channel with itself, averaging the PAC values:
       ```python
       pac_features = compute_pac_connectivity(data, method='tort', mode='self', approach_pac='mean')
       ```
    """

    n_channel, n_times = data.shape
    low_fq_width = band[0, 1] - band[0, 0]
    high_fq_width = band[1, 1] - band[1, 0]
    low_fq_range =np.linspace(band[0,0],band[0,1],low_fq_width+1)
    high_fq_range = np.linspace(band[1, 0], band[1, 1], high_fq_width+1)
    c = Comodulogram(fs=sfreq, low_fq_range=low_fq_range, high_fq_range=high_fq_range,method=method,
                     n_surrogates=n_surrogates,n_jobs=-1)
    if mode== "self":
        feature=np.zeros((n_channel))
        for N_channel in range(n_channel):
            sig = data[N_channel, :]
            pac_matrix=c.fit(sig).comod_
            if approach_pac=="mean":
                  feature[N_channel] = np.mean(pac_matrix)
            elif approach_pac=="max":
                feature[N_channel] = np.max(pac_matrix)

    elif mode== "non-self":
        feature = np.zeros((n_channel,n_channel))
        for i_channel in range(n_channel):
            for j_channel in range(n_channel):
                sig_low = data[i_channel, :]
                sig_high = data[j_channel, :]
                pac_matrix = c.fit(sig_low,sig_high).comod_
                if approach_pac == "mean":
                    feature[i_channel,j_channel] =  np.mean(pac_matrix)
                elif approach_pac == "max":
                    feature[i_channel, j_channel] = np.max(pac_matrix)
    feature = feature.reshape(-1)
    return feature

def compute_correlation_dimension(data,emb_dim=10):
    """
    Args:
        data:        ndarray,           shape (n_channels, n_times)
        emb_dim:     int                嵌入维度默认为10
    Returns:         feature            shape (n_channels)
    """
    import nolds
    n_channel, n_times = data.shape
    feature = np.zeros((n_channel))
    for i_channel in  range(n_channel):
        feature[i_channel] = nolds.corr_dim(data[i_channel,:],emb_dim)
    return feature
def compute_dispersion_entropy(data,classes=10,scale=1,emb_dim=2,delay=1,
                                   mapping_type='cdf',de_normalize=False, A=100,Mu=100,return_all=False,warns=True):
    """
    Args:
        data:        ndarray,           shape (n_channels, n_times)
        classes: number of classes - (levels of quantization of amplitude) (default=10)
        emb_dim: embedding dimension,
        delay  : time delay (default=1)
        scale  : downsampled signal with low resolution  (default=1)  - for multipscale dispersion entropy
        mapping_type: mapping method to discretizing signal (default='cdf')
               : options = {'cdf','a-law','mu-law','fd'}
        A  : factor for A-Law- if mapping_type = 'a-law'
        Mu : factor for μ-Law- if mapping_type = 'mu-law'

        de_normalize: (bool) if to normalize the entropy, to make it comparable with different signal with different
                     number of classes and embeding dimensions. default=0 (False) - no normalizations

        if de_normalize=1:
           - dispersion entropy is normalized by log(Npp); Npp=total possible patterns. This is classical
             way to normalize entropy since   max{H(x)}<=np.log(N) for possible outcomes. However, in case of
             limited length of signal (sequence), it would be not be possible to get all the possible patterns
             and might be incorrect to normalize by log(Npp), when len(x)<Npp or len(x)<classes**emb_dim.
             For example, given signal x with discretized length of 2048 samples, if classes=10 and emb_dim=4,
             the number of possible patterns Npp = 10000, which can never be found in sequence length < 10000+4.
             To fix this, the alternative way to nomalize is recommended as follow.
           - select this when classes**emb_dim < (N-(emb_dim-1)*delay)

          de_normalize=2: (recommended for classes**emb_dim > len(x)/scale)
           - dispersion entropy is normalized by log(Npf); Npf [= (len(x)-(emb_dim - 1) * delay)]
             the total  number of patterns founds in given sequence. This is much better normalizing factor.
             In worst case (lack of better word) - for a very random signal, all Npf patterns could be different
             and unique, achieving the maximum entropy and for a constant signal, all Npf will be same achieving to
             zero entropy
           - select this when classes**emb_dim > (N-(emb_dim-1)*delay)

          de_normalize=3:
           - dispersion entropy is normalized by log(Nup); number of total unique patterns (NOT RECOMMENDED)
             -  it does not make sense (not to me, at least)

          de_normalize=4:
           - auto select normalizing factor
           - if classes**emb_dim > (N-(emb_dim-1)*delay), then de_normalize=2
           - if classes**emb_dim > (N-(emb_dim-1)*delay), then de_normalize=2
    Returns:         feature            shape (n_channels)
    """
    import spkit
    n_channel, n_times = data.shape
    feature = np.zeros((n_channel))
    for i_channel in  range(n_channel):
          feature[i_channel],_ = spkit.dispersion_entropy(data[i_channel,:],classes,scale,emb_dim,delay,
                                   mapping_type,de_normalize, A,Mu,return_all,warns)
    return feature
def compute_aperiodic_periodic_offset_exponent_cf(data, sfreq=250, n=1024,freq_range=None, method="welch"):
    """
    Compute aperiodic and periodic parameters of the power spectrum from EEG data.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        sfreq (int): Sampling frequency of the signal. Default is 250 Hz.
        n (int): Number of frequency points for the Fourier transform. Typically set to the number of data points.
        freq_range (list or None): Frequency range for fitting with FOOOF. Default is None.
        method (str): Method for computing the power spectrum. Options are:
            - "fft": Fast Fourier Transform
            - "welch": Welch's method

    Returns:
        ndarray: Flattened array of aperiodic and periodic parameters with shape (n_channels * 2,).

    Notes:
        - Computes the offset and exponent of the aperiodic component and the periodic component of the power spectrum.
        - Uses the FOOOFGroup for fitting the power spectrum.
    """

    n_channel, n_times = data.shape
    feature = np.zeros((n_channel,2))
    if method =="fft":
        freqs = np.fft.fftfreq(n_times, 1 / sfreq)
        spectrum_frequencies = np.abs(np.fft.fft(data,n=n_times))
    elif method =="welch":
        freqs, spectrum_frequencies = welch(data, sfreq, nperseg=sfreq)
    spectrum_frequencies[:, freqs >= 0]
    fg1 = FOOOFGroup(verbose = False)
    fg1.fit(freqs, np.square(spectrum_frequencies)/(n**2), freq_range=freq_range)
    #[Offset, Exponent].
    feature[:,:2] = fg1.get_params("aperiodic_params")
    #[CF, PW, BW]..
    # peak_para=fg1.get_params('peak_params')
    # for i_channel in range(n_channel):
    #     idx_i_channel = np.where(peak_para[:,3] == i_channel)[0]
    #     peak_i_channel = peak_para[idx_i_channel,:]
    #     feature[i_channel,2] = peak_i_channel[np.argmax(peak_i_channel[:,1]),0]
    # bands = Bands({'alpha': [8, 12]})
    # feature[:,3] = get_band_peak_fg(fg1, bands.alpha)[:,0]
    feature = feature.T.reshape(-1)
    return feature

def compute_offset_exponent_cf(data,sfreq=250,n=1024,freq_range=None):
    """
    Compute the offset and exponent of the power spectrum from EEG data.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        sfreq (int): Sampling frequency of the signal. Default is 250 Hz.
        n (int): Number of frequency points for the Fourier transform. Typically set to the number of data points.

    Returns:
        ndarray: Flattened array of offset and exponent parameters with shape (n_channels * 2,).

    Notes:
        - Computes the median frequency and the spectral slope (offset and exponent) using `compute_spect_slope`.
        - The spectral slope is inverted in the returned feature array.
    """
    from mne_features.univariate import compute_spect_slope
    n_channel, n_times = data.shape
    feature = np.zeros((n_channel,2))
    feature[:,2]=compute_Median_Frequency(data,sfreq=sfreq,band=np.array([[8,12]]),N=n)
    # print(feature[:,2])
    if freq_range is None:
        freq_range=[0.1,50]
    slope_para=compute_spect_slope(sfreq, data, fmin=freq_range[0], fmax=freq_range[1],with_intercept=True, psd_method='welch', psd_params=None)
    slope_para=slope_para.reshape(n_channel,4)
    # print(slope_para)
    # print(slope_para.shape)
    # intercept, slope,
    feature[:, :2]=slope_para[:, :2]
    feature = feature.T.reshape(-1)
    return feature

##测试

def calculate_channel_correlation_pearson(channel_data1, channel_data2):
    pearson = (np.corrcoef(channel_data1, channel_data2))[0, 1]
    return np.abs(pearson)


def calculate_channel_correlation_plv(channel_data1, channel_data2):
    # 使用希尔伯特变换提取相位
    analytic_signal1 = hilbert(channel_data1)
    analytic_signal2 = hilbert(channel_data2)
    phase1 = np.angle(analytic_signal1)
    phase2 = np.angle(analytic_signal2)
    # 计算 PLV
    phase_diff = phase1 - phase2
    plv = np.abs(np.mean(np.exp(1j * phase_diff)))
    return plv


def calculate_channel_correlation_coh(channel_data1, channel_data2,sfreq):
    # 计算每对信号之间的相干性
    _, coh = coherence(channel_data1, channel_data2, fs=sfreq, nperseg=channel_data1.shape[0] // 2)
    # 取相干性矩阵的平均值作为代表值
    mean_coh = np.mean(coh)
    return mean_coh


##测试

def get_power_from_channel(data,wind,windowsover,i_channel,channel,sfreq,freq1,freq2):
    """
    Compute the power of a specified frequency band for a given channel.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        wind (int): Window length for the spectrogram.
        windowsover (int): Number of overlapping windows for the spectrogram.
        i_channel (int): Index of the channel to analyze.
        channel (list): List of channel names.
        sfreq (int): Sampling frequency of the signal.
        freq1 (float): Lower bound of the frequency range.
        freq2 (float): Upper bound of the frequency range.

    Returns:
        ndarray: Power of the specified frequency band for the given channel.

    Notes:
        - Uses `plt.specgram` to compute the power spectrum of the channel data.
    """
    L = channel.index(i_channel)
    signL = data[L, :]
    powerL, freqsL, t, _ = plt.specgram(signL, mode='psd', NFFT=wind, noverlap=windowsover, Fs=sfreq)
    indFreqs = [i for i, x in enumerate(freqsL) if freq1 <= x <= freq2]
    pow_L = powerL[indFreqs, 0]
    return pow_L

def compute_alpha_asymetry(data, sfreq=100, freq1=8, freq2=12, left='F3', right='F4', channel=None, mode="eeglab"):
    """
    Compute the alpha asymmetry between two specified channels.

    Args:
        data (ndarray): Input data with shape (n_channels, n_times).
        sfreq (int): Sampling frequency of the signal. Default is 100 Hz.
        freq1 (float): Lower bound of the alpha frequency range.
        freq2 (float): Upper bound of the alpha frequency range.
        left (str): Name of the left channel.
        right (str): Name of the right channel.
        channel (list or None): List of channel names. Default is None, in which case a default list is used.
        mode (str): Method for computing alpha asymmetry. Options are:
            - "eeglab": Method used in EEGLAB.
            - "definition_ln": Logarithmic difference.
            - "definition_ratio": Ratio difference.
            - "definition_lnratio": Logarithmic ratio difference.
            - "definition_ln_rel": Logarithmic relative difference.
            - "definition_ratio_rel": Ratio relative difference.
            - "definition_lnratio_rel": Logarithmic ratio relative difference.

    Returns:
        ndarray: Array of alpha asymmetry values with shape (n_channels,).

    Notes:
        - Computes alpha asymmetry using different methods depending on the `mode` parameter.
        - If `mode` is "eeglab", uses the power spectral density (PSD) of the specified channels.
    """
    if channel is None:
        channel = ['Fp1', 'Fp2', 'F7', 'F3', 'Fz', 'F4', 'F8', 'FC3', 'FCz', 'FC4',
                   'T7', 'C3', 'Cz', 'C4', 'T8', 'CP3', 'CPz', 'CP4', 'P7', 'P3',
                   'Pz', 'P4', 'P8', 'O1', 'Oz', 'O2']
    n_channel, length = data.shape
    try:
        if mode =="eeglab":
                wind = int(np.floor(length / 1.5))
                windowsover = int(np.floor(length / 1.5 / 2))
                pow_L = get_power_from_channel(data,wind,windowsover,left,channel,sfreq,freq1,freq2)
                pow_R = get_power_from_channel(data, wind, windowsover, right, channel,sfreq,freq1,freq2)
                FAA = np.mean(np.abs(np.log(pow_R) - np.log(pow_L)))
        elif "definition" in mode:
            info = mne.create_info(channel, sfreq)
            raw = mne.io.RawArray(data, info)
            psd = raw.compute_psd(method="welch", fmin=0.5, fmax=sfreq / 2, picks="all")
            alpha_mask = ((psd.freqs >= freq1) & (psd.freqs <= freq2))

            F3 = psd.copy().pick([left])
            F4 = psd.copy().pick([right])
            F3_total = F3.get_data([left]).squeeze()
            F4_total = F4.get_data([right]).squeeze()

            F3_alpha = F3_total[alpha_mask]
            F4_alpha = F4_total[alpha_mask]

            F3_rel_alpha = F3_alpha.mean() / F3_total.mean()
            F4_rel_alpha = F4_alpha.mean() / F4_total.mean()
            if mode == "definition_ln":
                FAA = np.mean(np.abs(np.log(F4_alpha) - np.log(F3_alpha)))
            elif mode == "definition_ratio":
                FAA = np.mean(np.abs((F4_alpha - F3_alpha) / (F3_alpha + F4_alpha)))
            elif mode == "definition_lnratio":
                FAA = np.mean(np.abs((np.log(F4_alpha) - np.log(F3_alpha)) / (np.log(F3_alpha) + np.log(F4_alpha))))
            elif mode == "definition_ln_rel":
                FAA = np.abs(np.log(F4_rel_alpha) - np.log(F3_rel_alpha))
            elif mode == "definition_ratio_rel":
                FAA = np.abs((F4_rel_alpha - F3_rel_alpha) / (F3_rel_alpha + F4_rel_alpha))
            elif mode == "definition_lnratio_rel":
                FAA = np.abs(
                    (np.log(F4_rel_alpha) - np.log(F3_rel_alpha)) / (np.log(F3_rel_alpha) + np.log(F4_rel_alpha)))
        feature = np.zeros(n_channel) + FAA
    except:
        feature = np.zeros(n_channel)
    return feature
def compute_pow_freq_bands_remove_aperiodic(data, sfreq=250, freq_bands=np.array([[1, 4], [4, 8], [8, 12], [12, 30], [30, 40]]),
                                           psd_method='welch', log=False, freq_range=None):
    """
    Compute the power in specified frequency bands after removing aperiodic components.

    Args:
        data (ndarray): Shape (n_channels, n_times). The input signal data.
        sfreq (int): The sampling frequency of the signal.
        freq_bands (ndarray): An array of frequency bands for power computation. Shape (n_bands, 2).
        psd_method (str): The method to use for computing the power spectral density ('fft' or 'welch').
        log (bool): Whether to apply a logarithm to the resulting power values.
        freq_range (tuple or None): The frequency range for aperiodic component fitting. If None, no range is used.

    Returns:
        ndarray: The power in each frequency band after removing aperiodic components. Flattened array of shape (n_channels * n_bands,).
    """
    n_channel, n_times = data.shape
    if psd_method == "fft":
        freqs = np.fft.fftfreq(n_times, 1/sfreq)
        spectrum_frequencies = np.abs(np.fft.fft(data, n=n_times))
    elif psd_method == "welch":
        freqs, spectrum_frequencies = welch(data, sfreq, nperseg=sfreq)
    spectrum_frequencies[:,freqs >= 0]
    fg1 = FOOOFGroup(verbose=False)
    fg1.fit(freqs, np.square(spectrum_frequencies)/(sfreq*n_times), freq_range=freq_range)
    offset_ex_list = fg1.get_params('aperiodic_params')
    aperiodic_value = np.zeros_like(spectrum_frequencies)
    for i_channel in range(n_channel):
        for idx in np.where(freqs > 0)[0]:
            aperiodic_value[i_channel, idx] = np.power(10, offset_ex_list[i_channel, 0] -
                                       offset_ex_list[i_channel, 1] * np.log10(freqs[idx]))
    all_psd = pow_freq_bands_from_spectrum(freq_bands, freqs, spectrum_frequencies)
    ap_psd =  pow_freq_bands_from_spectrum(freq_bands, freqs, aperiodic_value)
    remaind_psd = all_psd-ap_psd
    if log:
        remaind_psd = np.log10(remaind_psd)
    return remaind_psd.T.reshape(-1)
def compute_pow_freq_bands_cd(data, sfreq=250, freq_bands=np.array([[1, 4], [4, 8], [8, 12], [12, 30], [30, 40]]),
                                           psd_method='welch', log=False):
    """
    Compute the power in specified frequency bands from the power spectral density.

    Args:
        data (ndarray): Shape (n_channels, n_times). The input signal data.
        sfreq (int): The sampling frequency of the signal.
        freq_bands (ndarray): An array of frequency bands for power computation. Shape (n_bands, 2).
        psd_method (str): The method to use for computing the power spectral density ('fft' or 'welch').
        log (bool): Whether to apply a logarithm to the resulting power values.

    Returns:
        ndarray: The power in each frequency band. Flattened array of shape (n_channels * n_bands,).
    """
    n_channel, n_times = data.shape
    if psd_method == "fft":
        freqs = np.fft.fftfreq(n_times, 1/sfreq)
        spectrum_frequencies = np.abs(np.fft.fft(data, n=n_times))
    elif psd_method == "welch":
        freqs, spectrum_frequencies = welch(data, sfreq, nperseg=sfreq)
    spectrum_frequencies[:,freqs >= 0]
    all_psd = pow_freq_bands_from_spectrum(freq_bands, freqs, spectrum_frequencies)
    if log:
        all_psd = np.log10(all_psd)
    return all_psd.T.reshape(-1)

def pow_freq_bands_from_spectrum(freq_bands,freqs,spectrum_frequencies):
    """
    Compute the power in specified frequency bands from the power spectral density.

    Args:
        freq_bands (ndarray): An array of frequency bands for power computation. Shape (n_bands, 2).
        freqs (ndarray): Array of frequency values corresponding to the power spectrum.
        spectrum_frequencies (ndarray): The power spectrum of the data. Shape (n_channels, n_freqs).

    Returns:
        ndarray: The power in each frequency band. Shape (n_channels, n_bands).
    """
    n_bands = freq_bands.shape[0]
    n_channels = spectrum_frequencies.shape[0]
    feature = np.zeros((n_channels, n_bands))
    for index, i_band in enumerate(freq_bands):
        feature[:,index] = np.sum(spectrum_frequencies[:,np.where((freqs >= i_band[0]) & (freqs <= i_band[1]))[0]], axis=1)
    return feature



def compute_periodic_pac_connectivity(data, sfreq=250, n=1024, method="tort",
                                      band=np.array([[4, 8],[30,45]]), n_surrogates=0, mode="self", approach_pac="mean"):
    """
    Compute periodic phase-amplitude coupling (PAC) connectivity from the signal data.

    Args:
        data (ndarray): Shape (n_channels, n_times). The input signal data.
        sfreq (int): The sampling frequency of the signal.
        n (int): The number of points for Fourier transform.
        method (str): The method to use for PAC computation ('tort' or others).
        band (ndarray): An array specifying the frequency bands for PAC computation. Shape (2, 2).
        n_surrogates (int): The number of surrogate data to compute for significance testing.
        mode (str): The mode of PAC computation ('self' or others).
        approach_pac (str): The approach to compute PAC ('mean' or others).

    Returns:
        ndarray: The PAC connectivity feature. Flattened array of shape (n_channels * n_channels,).
    """

    import matplotlib.pyplot as plt
    from scipy.interpolate import interp1d
    n_channel, n_times = data.shape
    n=n_times
    spectrum_frequencies = np.zeros((n_channel,n))
    spectrum_phase = np.zeros((n_channel,n))
    reconstructed_signal = np.zeros_like(data)
    freqs = np.fft.fftfreq(n, 1 / sfreq)
    for i_channel in range(n_channel):
        fft_result = np.fft.fft(data[i_channel,:],n=n)
        spectrum_frequencies[i_channel, :]=np.abs(fft_result)
        spectrum_phase[i_channel,:] = np.angle(fft_result)
    fg1 = FOOOFGroup(verbose=False)
    fg1.fit(freqs, spectrum_frequencies)

    F=freqs
    gaussian_values = np.zeros_like(spectrum_frequencies)
    aps1 = fg1.get_params('gaussian_params')
    #print(aps1)
    gaussian_list=aps1[:,3]
    for i_channel in range(n_channel):
         i_channel_dx=np.where(gaussian_list==i_channel)[0]
         for idx in i_channel_dx:
            gaussian_values[i_channel,:] = aps1[idx,1] * norm.pdf(F, aps1[idx,0], aps1[idx,2])+gaussian_values[i_channel,:]
    gaussian_values = np.power(10, gaussian_values)
    for idx in np.where(freqs < 0)[0]:
        fre__ = -freqs[idx]
        try:
            gaussian_values[:, idx] = gaussian_values[:, np.where(freqs == fre__)[0][0]]
        except:
            print(freqs[idx])


    for i_channel in range(n_channel):
        # amplitude_interp = interp1d(freqs, gaussian_values[i_channel,:], kind='linear', fill_value='extrapolate')
        # phase_interp = interp1d(freqs, spectrum_phase[i_channel,:], kind='linear', fill_value='extrapolate')
        # new_sampling_rate = n_times / (gaussian_values[i_channel,:].shape[0] - 1) * sfreq
        # new_freq_vector = np.fft.fftfreq(n_times, d=1 / new_sampling_rate)
        #
        # # 插值得到新的幅值信号和相位信号
        # new_amplitude = amplitude_interp(new_freq_vector)
        # new_phase = phase_interp(new_freq_vector)
        # reconstructed_signal[i_channel, :] = np.fft.ifft(
        #     new_amplitude * np.exp(1j * new_phase))

        reconstructed_signal[i_channel,:] = np.fft.ifft(gaussian_values[i_channel,:] * np.exp(1j * spectrum_phase[i_channel,:]))

    # plt.plot(reconstructed_signal[i_channel,:])
    # plt.xlabel('time')
    # plt.ylabel('reconstructed_signal')
    # plt.show()


    feature=compute_pac_connectivity(reconstructed_signal,sfreq=sfreq, method=method, band=band, n_surrogates=n_surrogates,mode=mode,approach_pac=approach_pac)
    return feature
    
   
    
    










#!/usr/bin/env python

from __future__ import division

from ..tools.tools import *


def cum3est(y, maxlag, nsamp, overlap, flag, k1):
    """
  UM3EST Third-order cumulants.
  Should be invoked via "CUMEST" for proper parameter checks
  Parameters:
           y: input data vector (column)
      maxlag: maximum lag to be computed
    samp_seg: samples per segment
     overlap: percentage overlap of segments
       flag : 'biased', biased estimates are computed  [default]
              'unbiased', unbiased estimates are computed.
          k1: the fixed lag in c3(m,k1): see below

  Output:
       y_cum:  estimated third-order cumulant,
               C3(m,k1)  -maxlag <= m <= maxlag
  """

    (n1, n2) = np.shape(y)
    N = n1 * n2
    minlag = -maxlag
    overlap = np.fix(overlap / 100 * nsamp)
    nrecord = np.fix((N - overlap) / (nsamp - overlap))
    nadvance = nsamp - overlap

    y_cum = np.zeros([maxlag - minlag + 1, 1])

    ind = np.arange(nsamp).T
    nlags = 2 * maxlag + 1
    zlag = maxlag
    if flag == 'biased':
        scale = np.ones([nlags, 1]) / nsamp
    else:
        lsamp = nsamp - abs(k1)
        scale = make_arr((range(lsamp - maxlag, lsamp + 1), range(lsamp - 1, lsamp - maxlag - 1, -1)), axis=1).T
        (m2, n2) = scale.shape
        scale = np.ones([m2, n2]) / scale

    y = y.ravel(order='F')
    for i in np.arange(nrecord):
        x = y[ind]
        x = x - np.mean(x)
        cx = np.conj(x)
        z = x * 0

        # create the "IV" matrix: offset for second lag
        if k1 > 0:
            z[0:nsamp - k1] = x[0:nsamp - k1] * cx[k1:nsamp]
        else:
            z[-k1:nsamp] = x[-k1:nsamp] * cx[0:nsamp + k1]

        # compute third-order cumulants
        y_cum[zlag] = y_cum[zlag] + np.dot(z.T, x)

        for k in np.arange(1, maxlag + 1):
            y_cum[zlag - k] = y_cum[zlag - k] + np.dot(z[k:nsamp].T, x[0:nsamp - k])
            y_cum[zlag + k] = y_cum[zlag + k] + np.dot(z[0:nsamp - k].T, x[k:nsamp])

        ind = ind + int(nadvance)

    y_cum = y_cum * scale / nrecord

    return y_cum



#!/usr/bin/env python

from __future__ import division

import matplotlib.pyplot as plt
import scipy.io as sio

from ..tools.tools import *

np.set_printoptions(linewidth=120)


def bispectrumi(y, nlag=None, nsamp=None, overlap=None,
                flag='biased', nfft=None, wind=None):
    """
  Parameters:
    y       - data vector or time-series
    nlag    - number of lags to compute [must be specified]
    segsamp - samples per segment    [default: row dimension of y]
    overlap - percentage overlap     [default = 0]
    flag    - 'biased' or 'unbiased' [default is 'unbiased']
    nfft    - FFT length to use      [default = 128]
    wind    - window function to apply:
              if wind=0, the Parzen window is applied (default)
              otherwise the hexagonal window with unity values is applied.

  Output:
    Bspec   - estimated bispectrum  it is an nfft x nfft array
              with origin at the center, and axes pointing down and to the right
    waxis   - frequency-domain axis associated with the bispectrum.
            - the i-th row (or column) of Bspec corresponds to f1 (or f2)
              value of waxis(i).
  """

    (ly, nrecs) = y.shape
    if ly == 1:
        y = y.reshape(1, -1)
        ly = nrecs
        nrecs = 1

    if not overlap: overlap = 0
    overlap = min(99, max(overlap, 0))
    if nrecs > 1: overlap = 0
    if not nsamp: nsamp = ly
    if nsamp > ly or nsamp <= 0: nsamp = ly
    if not 'flag': flag = 'biased'
    if not nfft: nfft = 128
    if not wind: wind = 0

    nlag = min(nlag, nsamp - 1)
    if nfft < 2 * nlag + 1:
        nfft = 2 ^ nextpow2(nsamp)

    # create the lag window
    Bspec = np.zeros([nfft, nfft])
    if wind == 0:
        indx = np.array([range(1, nlag + 1)]).T
        window = make_arr((1, np.sin(np.pi * indx / nlag) / (np.pi * indx / nlag)), axis=0)
    else:
        window = np.ones([nlag + 1, 1])
    window = make_arr((window, np.zeros([nlag, 1])), axis=0)

    # cumulants in non-redundant region
    overlap = np.fix(nsamp * overlap / 100)
    nadvance = nsamp - overlap
    nrecord = np.fix((ly * nrecs - overlap) / nadvance)

    c3 = np.zeros([nlag + 1, nlag + 1])
    ind = np.arange(nsamp)
    y = y.ravel(order='F')

    s = 0
    for k in np.arange(nrecord):
        x = y[ind].ravel(order='F')
        x = x - np.mean(x)
        ind = ind + int(nadvance)

        for j in np.arange(nlag + 1):
            z = x[range(nsamp - j)] * x[range(j, nsamp)]
            for i in np.arange(j, nlag + 1):
                Sum = np.dot(z[range(nsamp - i)].T, x[range(i, nsamp)])
                if flag == 'biased':
                    Sum = Sum / nsamp
                else:
                    Sum = Sum / (nsamp - i)
                c3[i, j] = c3[i, j] + Sum

    c3 = c3 / nrecord

    # cumulants elsewhere by symmetry
    c3 = c3 + np.tril(c3, -1).T  # complete I quadrant
    c31 = c3[1:nlag + 1, 1:nlag + 1]
    c32 = np.zeros([nlag, nlag])
    c33 = np.zeros([nlag, nlag])
    c34 = np.zeros([nlag, nlag])
    for i in np.arange(nlag):
        x = c31[i:nlag, i]
        c32[nlag - 1 - i, 0:nlag - i] = x.T
        c34[0:nlag - i, nlag - 1 - i] = x
        if i + 1 < nlag:
            x = np.flipud(x[1:len(x)])
            c33 = c33 + np.diag(x, i + 1) + np.diag(x, -(i + 1))

    c33 = c33 + np.diag(c3[0, nlag:0:-1])

    cmat = make_arr(
        (make_arr((c33, c32, np.zeros([nlag, 1])), axis=1),
         make_arr((make_arr((c34, np.zeros([1, nlag])), axis=0), c3), axis=1)),
        axis=0
    )

    # apply lag-domain window
    wcmat = cmat
    if wind != -1:
        indx = np.arange(-1 * nlag, nlag + 1).T
        window = window.reshape(-1, 1)
        for k in np.arange(-nlag, nlag + 1):
            wcmat[:, k + nlag] = (cmat[:, k + nlag].reshape(-1, 1) * \
                                  window[abs(indx - k)] * \
                                  window[abs(indx)] * \
                                  window[abs(k)]).reshape(-1, )

    # compute 2d-fft, and shift and rotate for proper orientation
    Bspec = np.fft.fft2(wcmat, (nfft, nfft))
    Bspec = np.fft.fftshift(Bspec)  # axes d and r; orig at ctr

    if nfft % 2 == 0:
        waxis = np.transpose(np.arange(-1 * nfft / 2, nfft / 2)) / nfft
    else:
        waxis = np.transpose(np.arange(-1 * (nfft - 1) / 2, (nfft - 1) / 2 + 1)) / nfft

    cont = plt.contourf(waxis, waxis, abs(Bspec), 100, cmap=plt.cm.Spectral_r)
    plt.colorbar(cont)
    plt.title('Bispectrum estimated via the indirect method')
    plt.xlabel('f1')
    plt.ylabel('f2')
    plt.show()

    return (Bspec, waxis)


def test():
    qpc = sio.loadmat(here(__file__) + '/demo/qpc.mat')
    dbic = bispectrumi(qpc['zmat'], 21, 64, 0, 'unbiased', 128, 1)


if __name__ == '__main__':
    test()



#!/usr/bin/env python

from __future__ import division

from .cum2est import *
from .cum2x import *


def cum4est(y, maxlag=0, nsamp=0, overlap=0, flag='biased', k1=0, k2=0):
  """
  CUM4EST Fourth-order cumulants.
  Parameters:
          Should be invoked via CUMEST for proper parameter checks
          y_cum = cum4est (y, maxlag, samp_seg, overlap, flag, k1, k2)
          Computes sample estimates of fourth-order cumulants
          via the overlapped segment method.
          y_cum = cum4est (y, maxlag, samp_seg, overlap, flag, k1, k2)
                 y: input data vector (column)
            maxlag: maximum lag
          samp_seg: samples per segment
           overlap: percentage overlap of segments
             flag : 'biased', biased estimates are computed
                  : 'unbiased', unbiased estimates are computed.
      k1,k2 : the fixed lags in C3(m,k1) or C4(m,k1,k2)

  Output:
      y_cum : estimated fourth-order cumulant slice
              C4(m,k1,k2)  -maxlag <= m <= maxlag
  """

  (n1, n2) = shape(y, 2)
  N = n1*n2
  overlap0 = overlap
  overlap = np.fix(overlap/100 * nsamp)
  nrecord = np.fix((N - overlap)/(nsamp - overlap))
  nadvance = nsamp - overlap

  # scale factors for unbiased estimates
  nlags = 2 * maxlag + 1
  zlag  = maxlag
  tmp   = np.zeros([nlags,1])
  if flag == 'biased':
    scale = np.ones([nlags,1])/nsamp
  else:
    ind = np.arange(-maxlag, maxlag+1).T
    kmin = min(0, min(k1, k2))
    kmax  = max(0,max(k1, k2))
    scale = nsamp - np.maximum(ind, kmax) + np.minimum(ind, kmin)
    scale = np.ones(nlags) / scale
    scale = scale.reshape(-1,1)

  mlag  = maxlag + max(abs(k1), abs(k2))
  mlag  = max(mlag, abs(k1-k2) )
  mlag1 = mlag + 1
  nlag  = maxlag
  m2k2  = np.zeros([2*maxlag+1,1])

  if np.any(np.any(np.imag(y) != 0)): complex_flag = 1
  else: complex_flag = 0

  # estimate second- and fourth-order moments combine
  y_cum  = np.zeros([2*maxlag+1, 1])
  R_yy   = np.zeros([2*mlag+1, 1])

  ind = np.arange(nsamp)
  for i in np.arange(nrecord):
    tmp = np.zeros([2*maxlag+1, 1])
    x = y[ind]
    x = x.ravel(order='F') - np.mean(x)
    z =  x * 0
    cx = np.conj(x)

    # create the "IV" matrix: offset for second lag
    if k1 >= 0:
      z[0:nsamp-k1] = x[0:nsamp-k1] * cx[k1:nsamp]
    else:
      z[-k1:nsamp] = x[-k1:nsamp] * cx[0:nsamp+k1]

    # create the "IV" matrix: offset for third lag
    if k2 >= 0:
      z[0:nsamp-k2] = z[0:nsamp-k2] * x[k2:nsamp]
      z[nsamp-k2:nsamp] = np.zeros([k2, 1])
    else:
      z[-k2:nsamp] = z[-k2:nsamp] * x[0:nsamp+k2]
      z[0:-k2] = np.zeros([-k2, 1])

    tmp[zlag] = tmp[zlag] + np.dot(z.T, x)

    for k in np.arange(1, maxlag+1):
      tmp[zlag-k] = tmp[zlag-k] + np.dot(z[k:nsamp].T, x[0:nsamp-k])
      tmp[zlag+k] = tmp[zlag+k] + np.dot(z[0:nsamp-k].T, x[k:nsamp])

    y_cum = y_cum + tmp * scale

    R_yy = cum2est(x, mlag, nsamp, overlap0, flag)
    #  We need E x(t)x(t+tau) stuff also:
    if complex_flag:
      M_yy = cum2x(np.conj(x), x, mlag, nsamp, overlap0, flag)
    else:
      M_yy = R_yy

    y_cum = y_cum - \
            R_yy[mlag1+k1-1] * R_yy[mlag1-k2-nlag-1:mlag1-k2+nlag] - \
            R_yy[k1-k2+mlag1-1] * R_yy[mlag1-nlag-1:mlag1+nlag] - \
            M_yy[mlag1+k2-1].T * M_yy[mlag1-k1-nlag-1:mlag1-k1+nlag]

    ind = ind + int(nadvance)


  y_cum = y_cum / nrecord

  return y_cum




#!/usr/bin/env python

from __future__ import division

import matplotlib.pyplot as plt
import numpy as np
import scipy.io as sio
from scipy.linalg import hankel

from ..tools.tools import nextpow2, flat_eq, here


def bicoherencex(w, x, y, nfft=None, wind=None, nsamp=None, overlap=None):
    """
  Direct (FD) method for estimating cross-bicoherence
  Parameters:
    w,x,y - data vector or time-series
          - should have identical dimensions
    nfft - fft length [default = power of two > nsamp]
           actual size used is power of two greater than 'nsamp'
    wind - specifies the time-domain window to be applied to each
           data segment; should be of length 'segsamp' (see below);
      otherwise, the default Hanning window is used.
    segsamp - samples per segment [default: such that we have 8 segments]
            - if x is a matrix, segsamp is set to the number of rows
    overlap - percentage overlap, 0 to 99  [default = 50]
            - if y is a matrix, overlap is set to 0.

  Output:
    bic     - estimated cross-bicoherence: an nfft x nfft array, with
              origin at center, and axes pointing down and to the right.
    waxis   - vector of frequencies associated with the rows and columns
              of bic;  sampling frequency is assumed to be 1.
  """

    if w.shape != x.shape or x.shape != y.shape:
        raise ValueError('w, x and y should have identical dimentions')

    (ly, nrecs) = y.shape
    if ly == 1:
        ly = nrecs
        nrecs = 1
        w = w.reshape(1, -1)
        x = x.reshape(1, -1)
        y = y.reshape(1, -1)

    if not nfft:
        nfft = 128

    if not overlap: overlap = 50
    overlap = max(0, min(overlap, 99))
    if nrecs > 1: overlap = 0
    if not nsamp: nsamp = 0
    if nrecs > 1: nsamp = ly
    if nrecs == 1 and nsamp <= 0:
        nsamp = np.fix(ly / (8 - 7 * overlap / 100))
    if nfft < nsamp:
        nfft = 2 ** nextpow2(nsamp)

    overlap = np.fix(overlap / 100 * nsamp)
    nadvance = nsamp - overlap
    nrecs = np.fix((ly * nrecs - overlap) / nadvance)

    if not wind:
        wind = np.hanning(nsamp)

    try:
        (rw, cw) = wind.shape
    except ValueError:
        (rw,) = wind.shape
        cw = 1

    if min(rw, cw) != 1 or max(rw, cw) != nsamp:
        print("Segment size is " + str(nsamp))
        print("Wind array is " + str(rw) + " by " + str(cw))
        print("Using default Hanning window")
        wind = np.hanning(nsamp)

    wind = wind.reshape(1, -1)

    # Accumulate triple products
    bic = np.zeros([nfft, nfft])
    Pyy = np.zeros([nfft, 1])
    Pww = np.zeros([nfft, 1])
    Pxx = np.zeros([nfft, 1])

    mask = hankel(np.arange(nfft), np.array([nfft - 1] + list(range(nfft - 1))))
    Yf12 = np.zeros([nfft, nfft])
    ind = np.transpose(np.arange(nsamp))
    w = w.ravel(order='F')
    x = x.ravel(order='F')
    y = y.ravel(order='F')

    for k,value in enumerate(nrecs):
        ws = w[ind]
        ws = (ws - np.mean(ws)) * wind
        Wf = np.fft.fft(ws, nfft) / nsamp
        CWf = np.conjugate(Wf)
        Pww = Pww + flat_eq(Pww, (Wf * CWf))

        xs = x[ind]
        xs = (xs - np.mean(xs)) * wind
        Xf = np.fft.fft(xs, nfft) / nsamp
        CXf = np.conjugate(Xf)
        Pxx = Pxx + flat_eq(Pxx, (Xf * CXf))

        ys = y[ind]
        ys = (ys - np.mean(ys)) * wind
        Yf = np.fft.fft(ys, nfft) / nsamp
        CYf = np.conjugate(Yf)
        Pyy = Pyy + flat_eq(Pyy, (Yf * CYf))

        Yf12 = flat_eq(Yf12, CYf.ravel(order='F')[mask])
        bic = bic + (Wf * np.transpose(Xf)) * Yf12

        ind = ind + int(nadvance)

    bic = bic / nrecs
    Pww = Pww / nrecs
    Pxx = Pxx / nrecs
    Pyy = Pyy / nrecs
    mask = flat_eq(mask, Pyy.ravel(order='F')[mask])

    bic = abs(bic) ** 2 / ((Pww * np.transpose(Pxx)) * mask)
    bic = np.fft.fftshift(bic)

    # Contour plot of magnitude bispectrum
    if nfft % 2 == 0:
        waxis = np.transpose(np.arange(-1 * nfft / 2, nfft / 2)) / nfft
    else:
        waxis = np.transpose(np.arange(-1 * (nfft - 1) / 2, (nfft - 1) / 2 + 1)) / nfft

    cont = plt.contourf(waxis, waxis, bic, 100, cmap=plt.cm.Spectral_r)
    plt.colorbar(cont)
    plt.title('Bicoherence estimated via the direct (FFT) method')
    plt.xlabel('f1')
    plt.ylabel('f2')

    colmax, row = bic.max(0), bic.argmax(0)
    maxval, col = colmax.max(0), colmax.argmax(0)
    print('Max: bic(' + str(waxis[col]) + ',' + str(waxis[col]) + ') = ' + str(maxval))
    plt.show()

    return bic, waxis


def test():
    nl1 = sio.loadmat(here(__file__) + '/demo/nl1.mat')
    dbic = bicoherencex(nl1['x'], nl1['x'], nl1['y'])


if __name__ == '__main__':
    test()



#!/usr/bin/env python

from __future__ import division

from .cum2x import *


def cum4x(w, x, y, z, maxlag=0, nsamp=0, overlap=0, flag='biased', k1=0, k2=0):
  """
  Fourth-order cross-cumulants.
  Parameters:
     w,x,y,z  - data vectors/matrices with identical dimensions
               if w,x,y,z are matrices, rather than vectors, columns are
               assumed to correspond to independent realizations,
               overlap is set to 0, and samp_seg to the row dimension.
      maxlag - maximum lag to be computed    [default = 0]
    samp_seg - samples per segment  [default = data_length]
     overlap - percentage overlap of segments [default = 0]
               overlap is clipped to the allowed range of [0,99].
       flag : 'biased', biased estimates are computed  [default]
              'unbiased', unbiased estimates are computed.
      k1,k2 : the fixed lags in C4(m,k1,k2) defaults to 0

  Output:
       y_cum:  estimated fourth-order cross cumulant,
             c4(t1,t2,t3) := cum( w^*(t), x(t+t1), y(t+t2), z^*(t+t3) )
  """

  (lx, nrecs) = w.shape
  if (lx, nrecs) != x.shape or (lx, nrecs) != y.shape or (lx, nrecs) != z.shape:
    raise ValueError('w,x,y,z should have identical dimensions')

  if lx == 1:
    lx = nrecs
    nrecs = 1

  if maxlag < 0: raise ValueError('"maxlag" must be non-negative ')
  if nrecs > 1: nsamp = lx
  if nsamp <= 0 or nsamp > lx: nsamp = lx

  if nrecs > 1: overlap = 0
  overlap = max(0,min(overlap,99))

  overlap0 = overlap
  overlap  = np.fix(overlap/100 * nsamp)
  nadvance = nsamp - overlap

  if nrecs == 1:
    nrecs = np.fix((lx - overlap)/nadvance)


  # scale factors for unbiased estimates
  nlags = 2 * maxlag + 1
  zlag = maxlag

  tmp = np.zeros([nlags, 1])
  if flag == 'biased':
    scale = np.ones([nlags, 1]) / nsamp
    sc1 = 1/nsamp
    sc2 = sc1
    sc12 = sc1
  else:
    ind   = np.arange(-maxlag, maxlag+1).T
    kmin  = min(0,min(k1,k2))
    kmax  = max(0,max(k1,k2))
    scale = nsamp - np.maximum(ind,kmax) + np.minimum(ind,kmin)
    scale = np.ones(nlags) / scale
    sc1  = 1 / (nsamp - abs(k1))
    sc2  = 1 / (nsamp - abs(k2))
    sc12 = 1 / (nsamp - abs(k1 - k2))
    scale = scale.reshape(-1,1)

  # estimate second- and fourth-order moments combine
  y_cum  = np.zeros([2*maxlag+1, 1])
  rind = np.arange(-maxlag, maxlag+1)
  ind = np.arange(nsamp)

  print(nrecs)
  for i in np.arange(nrecs):
    tmp = y_cum * 0
    R_zy   = 0
    R_wy = 0
    M_wz = 0
    ws = w[ind]
    ws = ws - np.mean(ws)
    xs = x[ind]
    xs = xs - np.mean(xs)
    ys = y[ind]
    ys = ys - np.mean(ys)
    cys = np.conj(ys)
    zs = z[ind]
    zs = zs - np.mean(zs)

    ziv = xs * 0

    # create the "IV" matrix: offset for second lag
    if k1 >= 0:
      ziv[0:nsamp-k1] = ws[0:nsamp-k1] * cys[k1:nsamp]
      R_wy = R_wy + np.dot(ws[0:nsamp-k1].T, ys[k1:nsamp])
    else:
      ziv[-k1:nsamp] = ws[-k1:nsamp] * cys[0:nsamp+k1]
      R_wy = R_wy + np.dot(ws[-k1:nsamp].T, ys[0:nsamp+k1])

    # create the "IV" matrix: offset for third lag
    if k2 > 2:
      ziv[0:nsamp-k2] = ziv[0:nsamp-k2] * zs[k2:nsamp]
      ziv[nsamp-k2:nsamp] = np.zeros([k2, 1])
      M_wz = M_wz + np.dot(ws[0:nsamp-k2].T, zs[k2:nsamp])
    else:
      ziv[-k2:nsamp] = ziv[-k2:nsamp] * zs[0:nsamp+k2]
      ziv[0:-k2] = np.zeros([-k2, 1])
      M_wz = M_wz + np.dot(ws[-k2:nsamp].T, zs[0:nsamp-k2])


    if k1-k2 >= 0:
      R_zy = R_zy + np.dot(zs[0:nsamp-k1+k2].T, ys[k1-k2:nsamp])
    else:
      R_zy = R_zy + np.dot(zs[-k1+k2:nsamp].T, ys[0:nsamp-k2+k1])

    tmp[zlag] = tmp[zlag] + np.dot(ziv.T, xs)
    for k in np.arange(1, maxlag+1):
      tmp[zlag-k] = tmp[zlag-k] + np.dot(ziv[k:nsamp].T, xs[0:nsamp-k])
      tmp[zlag+k] = tmp[zlag+k] + np.dot(ziv[0:nsamp-k].T, xs[k:nsamp])

    print(y_cum.shape)
    y_cum = y_cum + tmp * scale # fourth-order moment estimates done
    print(y_cum.shape)

    R_wx = cum2x(ws,      xs, maxlag,         nsamp, overlap0, flag)
    R_zx = cum2x(zs,      xs, maxlag+abs(k2), nsamp, overlap0, flag)
    M_yx = cum2x(cys,     xs, maxlag+abs(k1), nsamp, overlap0, flag)

    y_cum = y_cum - R_zy * R_wx * sc12 - \
            R_wy * R_zx[rind - k2 + maxlag + abs(k2)] * sc1 - \
            M_wz.T * M_yx[rind - k1 + maxlag + abs(k1)] * sc2

    ind = ind + int(nadvance)

  y_cum = y_cum / nrecs

  return y_cum



def test():

  y = sio.loadmat(here(__file__) + '/demo/ma1.mat')['y']

  # The right results are:
  #           "biased": [-0.52343  -0.43057   1.16651   3.21583   1.98088  -0.38022  -1.05836]
  #           "unbiased": [-0.53962  -0.43936   1.17829   3.21583   2.00089  -0.38798  -1.09109]
  print(cum4x(y, y, y, y, 3, 100, 0, "biased"))
  print(cum4x(y, y, y, y, 3, 100, 0, "unbiased"))


if __name__ == '__main__':
  test()






import setuptools
from setuptools import find_packages

# with open("README.md", "r") as fh:
#     long_description = fh.read()

setuptools.setup(
    name="scuteegfe",
    version="0.0.4",
    author="bkxcyu",
    author_email="bkxcyu@gmail.com",
    description="EEG Signal Feature Exacting, a wrapper of mne_features",
    # long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages('src'),
    package_dir={"": "src"},
    install_requires=[
        'requests',
        'importlib-metadata; python_version >= "3.10"',
        'einops',
        'mne_features',
        'EMD-signal',
        'scipy',
        'PyWavelets',
        'pyts',
        'antropy',
        'pyentrp',
        'tftb',
        'statsmodels',
        'scipy',
        'emd-signal',
        'pyentrp',
        'nolds',
        'pactools',
        'EntropyHub',
'pykalman',
'mne_connectivity',
'spkit>= 0.0.9.6.8',
'fooof',
'nilearn',
'pybispectra'
],
)



#!/usr/bin/env python

from __future__ import division

from ..tools.tools import *


def cum2est(y, maxlag, nsamp, overlap, flag):
    """
  CUM2EST Covariance function.
  Should be involed via "CUMEST" for proper parameter checks.
  Parameters:
           y: input data vector (column)
      maxlag: maximum lag to be computed
    samp_seg: samples per segment (<=0 means no segmentation)
     overlap: percentage overlap of segments
        flag: 'biased', biased estimates are computed
              'unbiased', unbiased estimates are computed.

  Output:
       y_cum: estimated covariance,
              C2(m)  -maxlag <= m <= maxlag
  """

    (n1, n2) = shape(y, 2)
    N = n1 * n2
    overlap = np.fix(overlap / 100 * nsamp)
    nrecord = np.fix((N - overlap) / (nsamp - overlap))
    nadvance = nsamp - overlap

    y_cum = np.zeros([maxlag + 1, 1])
    ind = np.arange(nsamp)
    y = y.ravel(order='F')

    for i in np.arange(nrecord):
        x = y[ind]
        x = x - np.mean(x)

        for k in np.arange(maxlag + 1):
            y_cum[k] = y_cum[k] + np.dot(x[0:nsamp - k].T, x[k:nsamp])

        ind = ind + int(nadvance)

    if flag == 'biased':
        y_cum = y_cum / (nsamp * nrecord)
    else:
        y_cum = y_cum / (nrecord * (nsamp - np.matrix(range(maxlag + 1)).T))
        y_cum = np.asarray(y_cum)

    if maxlag > 0:
        y_cum = make_arr([np.conj(y_cum[maxlag:0:-1]), y_cum], axis=0)

    return y_cum



from mne_features.feature_extraction import extract_features
import copy
from einops import rearrange
from ..features.any_feature import *
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from copy import deepcopy
import warnings
import numbers

class Feature:
    mne_defined_funcs = {'mean', 'variance', 'std', 'ptp_amp', 'skewness', 'kurtosis', 'rms', 'quantile',
                         'hurst_exp', 'app_entropy', 'samp_entropy', 'decorr_time', 'pow_freq_bands',
                         'hjorth_mobility_spect', 'hjorth_complexity_spect', 'hjorth_mobility',
                         'hjorth_complexity', 'higuchi_fd', 'katz_fd', 'zero_crossings', 'line_length',
                         'spect_slope', 'spect_entropy', 'svd_entropy', 'svd_fisher_info', 'energy_freq_bands',
                         'spect_edge_freq', 'wavelet_coef_energy', 'teager_kaiser_energy'}

    funcs_subset_no_spect_slope = {'mean', 'variance', 'std', 'ptp_amp', 'skewness', 'kurtosis', 'rms', 'quantile',
                                   'hurst_exp', 'app_entropy', 'samp_entropy', 'decorr_time', 'pow_freq_bands',
                                   'hjorth_mobility_spect', 'hjorth_complexity_spect', 'hjorth_mobility',
                                   'hjorth_complexity', 'higuchi_fd', 'katz_fd', 'zero_crossings', 'line_length',
                                   'spect_entropy', 'svd_entropy', 'svd_fisher_info', 'energy_freq_bands',
                                   'spect_edge_freq', 'wavelet_coef_energy', 'teager_kaiser_energy','aac_connectivity','pac_connectivity'}


    def __init__(self, data=None, sfreq=250, selected_funcs=None, funcs_params=None, n_jobs=1, ch_names=None,
                 return_as_df=False, log_teager_kaiser_energy=False):
        """
        Initialize the feature extractor.

        Args:
            data (ndarray, optional): Input data with shape (n_samples, n_channels, n_features). If None, only available features will be printed.
            sfreq (int, optional): Sampling frequency, default is 250 Hz.
            selected_funcs (list of str, optional): List of feature function names to compute. If None, default functions will be used.
            funcs_params (dict, optional): Parameters for feature functions. If None, default parameters will be used.
            n_jobs (int, optional): Number of processes to use, default is 1.
            ch_names (list of str, optional): List of channel names. If None, channel names will not be used.
            return_as_df (bool, optional): Whether to return features as a pandas.DataFrame. Default is False.
            log_teager_kaiser_energy (bool, optional): Whether to compute the logarithm of Teager-Kaiser energy. Default is False.

        Warns:
            UserWarning: Issued if sfreq is not equal to 250.

        Attributes:
            funcs (list of callable): List of feature functions to compute.
            feature_names_order (list of str): Order of feature names.
            example_data (ndarray): Example data with shape (1, 1).
            n_channel (int): Number of channels in the data.
            funcs_params (dict): Parameters for feature functions.
            __features_raw (ndarray): Raw computed features.
            __features (ndarray): Features rearranged for output.
            __features_fix (bool): Flag indicating if features have been fixed.
            __list_multi_feature (list of str): List of multi-feature names.

        """
        if data is None:
            print('available features:', self.mne_defined_funcs)
            self.__features = None
            return
        if sfreq!=250:
            warnings.warn("提案的函数sfreq需要再以参数的形式传入", UserWarning)
        self.__feature_names = None
        self.__features = None
        self.log_teager_kaiser_energy = log_teager_kaiser_energy
        funcs, feature_names_order = self.get_funcs(selected_funcs)
        self.funcs = funcs
        self.feature_names_order = feature_names_order
        self.example_data = data[0, 0][None, None]
        self.conn_example_data = data[0][None]
        self.n_channel = data.shape[1]
        self.funcs_params = funcs_params
        features = extract_features(data, sfreq, funcs, funcs_params, n_jobs, ch_names, return_as_df)
        if return_as_df:
            self.__features_raw = features

        self.__features_raw = features
        self.__features = rearrange(self.__features_raw, 'b (feature channel) -> b channel feature',
                                    channel=self.n_channel)
        self.__features_fix = False
        self.__list_multi_feature = ['teager_kaiser_energy0', 'spect_slope0',
                          'energy_freq_bands0', 'wavelet_coef_energy0', 'pow_freq_bands0']
    def __repr__(self):
        """
        Generate a string representation of the FeatureExtractor instance.

        If the features attribute is None, a prompt to input data is returned. Otherwise,
        the string representation includes the number of epochs, channels, and features,
        along with the feature names.

        Returns:
            str: String representation of the FeatureExtractor instance.
        """
        if self.features is None:
            return 'you should input the data'
        else:
            n_epochs, n_channels, n_features = self.features.shape
            return str(n_epochs) + '(epochs) x ' + str(n_channels) + '(channels) x ' + str(n_features) + '(features)' \
                   + '\nfeature names: ' + str(self.feature_names)

    def get_funcs(self, selected_funcs):
        """
         Retrieve custom feature decomposition functions based on selected function names.

         Args:
             selected_funcs (list of str or list of tuple): List of feature function names or tuples of
                 (function name, function reference) to retrieve.

         Returns:
             tuple: A tuple containing:
                 - funcs (set of tuple): A set of tuples where each tuple is (function name, function reference).
                 - feature_names_order (list of str): List of feature function names in the order they were provided.

         Raises:
             AttributeError: If the format of `selected_funcs` is not valid.

        """
        # 获取自定义的特征分解函数
        selected_funcs = list(selected_funcs)
        for i, each in enumerate(selected_funcs):
            if not {each}.issubset(self.mne_defined_funcs):
                selected_funcs[i] = (each, eval('compute_' + each))
        # 获取自定义的特征名
        feature_names_order = []
        funcs = set(selected_funcs)
        for each in funcs:
            if isinstance(each, tuple):
                f_name = each[0]
                assert isinstance(f_name, str)
                feature_names_order.append(each[0])
            elif isinstance(each, str):
                feature_names_order.append(each)
            else:
                raise AttributeError
        return funcs, feature_names_order

    def fix_missing(self):
        """
        Fix missing values in the features.

        This method uses the mean strategy to impute missing values. If the mean strategy fails,
        it falls back to using a constant strategy with a fill value of 0.

        Returns:
            FeatureExtractor: A new instance of FeatureExtractor with missing values imputed.

        """
        from sklearn.impute import SimpleImputer
        imp_mean = SimpleImputer(missing_values=np.nan, strategy='mean')

        feature = np.zeros_like(self.features)
        for i, each_epoch in enumerate(self.features):
            try:
                feature[i] = imp_mean.fit_transform(each_epoch)
            except Exception as e:
                print('Can not fix missing value using "mean" method, now try constant method ', e)
                imp_constant = SimpleImputer(missing_values=np.nan, strategy='constant', fill_value=0)
                feature[i] = imp_constant.fit_transform(each_epoch)

        n_F = copy.deepcopy(self)
        n_F.features = feature
        return n_F

    def reorder(self):
        """
        Reorder features alphabetically by their names.

        Returns:
            FeatureExtractor: A new instance of FeatureExtractor with features reordered alphabetically.
        """
        features = self.features
        feature_names = self.feature_names
        order = feature_names.argsort()
        n_F = copy.deepcopy(self)
        n_F.__features = features[:, :, order]
        n_F.__feature_names = feature_names[order]
        return n_F

    @property
    def features(self):
        """
        Get the features with possible multi-feature fixing.

        Returns:
            ndarray: The features array.
        """
        if self.__features_fix is True:
            return self.__features
        if np.any(np.isin(self.__list_multi_feature, self.feature_names)):
            self.fix_multi_feature(log = self.log_teager_kaiser_energy)
            self.__features_fix = True
        return self.__features

    @features.setter
    def features(self, features):
        """
        Set the features array.

        Args:
            features (ndarray): The new features array.
        """
        self.__features = features

    @property
    def feature_names(self):
        """
        Get the feature names.

        Returns:
            ndarray: Array of feature names.
        """
        if self.__feature_names is not None:
            return self.__feature_names
        feature_names = []
        feature_shapes = []
        for each_fea in self.feature_names_order:
            try:
                params = list(self.funcs_params.keys())
                matching_params = []
                for each_param in params:
                    match = each_param.startswith(each_fea)
                    if match:
                        matching_params.append(each_param)
                param_dict = {}
                for each_param in matching_params:
                    param = {each_param: self.funcs_params[each_param]}
                    param_dict.update(param)
            except Exception as e:
                param_dict = None
            if each_fea in ["pac_connectivity","correlation_matrix"]:
                fea_ = Feature(self.conn_example_data, selected_funcs={each_fea}, funcs_params=param_dict)
            else:
                fea_ = Feature(self.example_data, selected_funcs={each_fea}, funcs_params=param_dict)
            fea_shape = fea_.__features.shape[2]
            feature_names.append(each_fea)
            feature_shapes.append(fea_shape)

        feature_indexs = []
        for i, each_fea in enumerate(feature_names):
            if int(feature_shapes[i]) > 1:
                fea_sub_name = []
                for each in range(int(feature_shapes[i])):
                    fea_sub_name.append(each_fea + (str(each)))
                feature_indexs.extend(fea_sub_name)
            else:
                feature_indexs.extend([each_fea])
        self.__feature_names = np.array(feature_indexs)
        return self.__feature_names

    @feature_names.setter
    def feature_names(self, f_names):
        """
        Get the feature names.

        Returns:
            ndarray: Array of feature names.
        """
        self.__feature_names = f_names

    def get_data(self,n_sample_list=None):
        """
       Get the feature vector array.

       Args:
           n_sample_list (list, optional): List of sample indices to slice the features array. Defaults to None.

       Returns:
           ndarray: The original or sliced features array.
       """
        if n_sample_list is None:
              fea= self.features
              return fea
        else:
            dim = self.features.ndim
            fea = self.features[n_sample_list,:,:]
            if fea.ndim<dim:
                fea=fea[None,:]
            return fea

    def copy(self):
        """
       Copy the instance of Feature.

       Returns:
           Feature: A copy of the object.
       """
        Feature = deepcopy(self)
        return Feature

    def __getitem__(self, item):
        """
        Slice the Feature instance on the sample dimension.

        Args:
            item (slice): Slice object for indexing the features array.

        Returns:
            Feature: A new instance of Feature with sliced features.
        """
        cls = type(self)
        Feature = self.copy()
        Feature.features = self.get_data(item)
        return Feature




    @staticmethod
    def plot_feature_sns(Feature1, Feature2, ch_names, sub_type1='type1', sub_type2='type2'):
        """
        Plot features using seaborn.

        Args:
            Feature1 (Feature): First Feature instance.
            Feature2 (Feature): Second Feature instance.
            ch_names (list): List of channel names.
            sub_type1 (str, optional): Description of the first dataset type. Defaults to 'type1'.
            sub_type2 (str, optional): Description of the second dataset type. Defaults to 'type2'.
        """
        fea1 = Feature1.features
        fea2 = Feature2.features

        # 归一化
        fea_concat = np.concatenate([fea1, fea2], axis=0)
        std = fea_concat.std(axis=0)
        mean = fea_concat.mean(axis=0)
        std[std == 0] = 1e-20
        mean[np.isnan(mean)] = 0
        fea_concat = (fea_concat - mean[None, :, :]) / (std[None, :, :])
        fea1 = fea_concat[:fea1.shape[0]]
        fea2 = fea_concat[fea1.shape[0]:]

        # 构造DataFrame
        df1 = Feature1.feature_df2plot(fea1, Feature1.feature_names, ch_names, sub_type=sub_type1)
        df2 = Feature2.feature_df2plot(fea2, Feature2.feature_names, ch_names, sub_type=sub_type2)
        df2plot = pd.concat([df1, df2])

        # 画图
        for ch in ch_names:
            plt.figure(figsize=[100, 10])
            sns.boxplot(data=df2plot.query('Channel=="' + ch + '"'), x='features', y='Value', hue='type')
            plt.title(ch)
            plt.show()

    @staticmethod
    def feature_df2plot(features, feature_names, ch_names, sub_type='sub_type_1'):
        """
        Convert features to a DataFrame for plotting.

        This method organizes features, feature names, and channel names into a pandas DataFrame,
        which can be used for visualization purposes, particularly with seaborn.

        Args:
            features (ndarray): The feature array with shape (n_samples, n_channels, n_features).
            feature_names (list): List of feature names.
            ch_names (list): List of channel names.
            sub_type (str, optional): Description of the dataset type. Defaults to 'sub_type_1'.

        Returns:
            pandas.DataFrame: DataFrame containing the features, channel names, type, and values,
                              structured for plotting.
        """
        from einops import repeat, rearrange
        fea_names = repeat(np.array(feature_names), 'n_f -> (n_epoch n_ch n_f)', n_epoch=features.shape[0],
                           n_ch=features.shape[1])
        ch_names = repeat(np.array(ch_names), 'n_ch -> (n_epoch n_ch n_f)', n_epoch=features.shape[0],
                          n_f=features.shape[2])
        values = rearrange(features, 'n_epoch n_ch n_f -> (n_epoch n_ch n_f)')

        return pd.DataFrame({
            'features': fea_names,
            'Channel': ch_names,
            'type': sub_type,
            'Value': values
        })

    @staticmethod
    def ttest_feature(Feature1, Feature2, ch_names):
        """
        Run t-test and visualize p-values using a heatmap.

        This function performs a t-test between the features of two Feature instances and visualizes the p-values
        using a heatmap. The heatmap shows the -log10 of the p-values for better visualization.

        Args:
            Feature1 (Feature): An instance of the Feature class.
            Feature2 (Feature): Another instance of the Feature class.
            ch_names (list): List of channel names.

        Returns:
            tuple: A tuple containing the t-statistics and p-values.
        """
        assert Feature1.features is not None and Feature2.features is not None
        from scipy import stats

        # TODO 这里默认Feature1和Feature2的特征顺序一致，如果不一致，将会有问题
        assert (Feature1.feature_names == Feature2.feature_names).all() == True, \
            print('feature not match:', Feature1.feature_names, Feature2.features)
        sta, p = stats.ttest_ind(Feature1.features, Feature2.features)
        log10_p = np.log10(p)
        thresh = np.log10(0.05)
        nan_index = np.isnan(p)
        if nan_index.sum() != 0:
            p[nan_index] = p[~nan_index].mean(axis=0)

        plt.figure(figsize=(16, 4))
        sns.heatmap(log10_p, square=True, center=thresh, cmap='coolwarm', vmin=-4, vmax=0,
                    yticklabels=ch_names, xticklabels=Feature1.feature_names)
        return sta, p
    def fix_multi_feature(self,log=True):
        """
            Fix multi-feature data by rearranging it.

            This function rearranges the multi-feature data to ensure proper alignment and, optionally,
            applies a logarithmic transformation to the 'teager_kaiser_energy' feature.

            Args:
                log (bool, optional): If True, applies a logarithmic transformation to the 'teager_kaiser_energy' feature.
                                      Defaults to True.

            Returns:
                None
        """
        present_features = [feature for feature in self.__list_multi_feature if feature in self.feature_names]
        for multi_feature_name in present_features:
            multi_feature_name = multi_feature_name [:-1]
            multi_feature_name_names = [multi_feature_name + str(i) for i in
                                          range(np.char.startswith(self.feature_names, multi_feature_name).sum())]

            get_index = lambda source, target: np.argwhere(source == target)[0, 0]
            reorder = lambda source, target: [get_index(each, target) for each in source]
            _rearrange_ = lambda features: rearrange(
                rearrange(features, 'n_sub n_ch n_fea-> n_sub (n_fea n_ch)'),
                'n_sub (n_ch n_fea)-> n_sub n_ch n_fea', n_ch=self.n_channel)
            ind = reorder(multi_feature_name_names, self.feature_names)
            if log and multi_feature_name =='teager_kaiser_energy':
                self.__features[:, :, ind] = np.log10(_rearrange_(self.__features[:, :, ind]))
            else:
                self.__features[:, :, ind] = _rearrange_(self.__features[:, :, ind])

    @staticmethod
    def moving_average_filter(data, window_size):
        """
        Apply a moving average filter to the data.

        This function applies a moving average filter with a specified window size to smooth the data.

        Args:
            data (array-like): The input data to be filtered.
            window_size (int): The size of the moving window.

        Returns:
            ndarray: The filtered data.
        """
        filtered_data = []
        window = [0] * window_size  # Initialize a window of size `window_size` with zeros
        for i in range(len(data)):
            window.pop(0)  # Remove the oldest data point from the window
            window.append(data[i])  # Add the current data point to the window
            filtered_data.append(sum(window) / window_size)  # Calculate the average and add it to the filtered data
        filtered_data = np.array(filtered_data)
        return filtered_data

    @staticmethod
    def lsd_KalmanFilter(data, window_size):
        """
        Apply a Kalman Filter for smoothing data.

        This function uses a Kalman Filter to smooth the input data within a specified window size.

        Args:
            data (array-like): The input data to be smoothed.
            window_size (int): The size of the processing window.

        Returns:
            ndarray: The smoothed data.
        """
        from pykalman import KalmanFilter
        window_num = data.shape[0] // window_size
        smoothed_feature = []
        for i_window in range(window_num + 1):
            begin_idx = window_size * i_window
            end_idx = window_size * (i_window + 1)
            if begin_idx >= data.shape[0]:
                continue
            if end_idx > data.shape[0]:
                end_idx = data.shape[0]
            data_window = data[begin_idx:end_idx]
            transition_covariance = np.diag([0.1, 0.1])
            transition_covariance = 0.1
            observation_covariance = 0.001
            initial_state_mean = np.mean(data_window)
            initial_state_covariance = 1
            kf = KalmanFilter(transition_covariance=transition_covariance,
                              observation_covariance=observation_covariance,
                              initial_state_mean=initial_state_mean,
                              initial_state_covariance=initial_state_covariance)
            # Estimate the parameters using the EM algorithm
            kf = kf.em(data_window)
            estimated_A = kf.transition_matrices
            estimated_C = kf.observation_matrices
            estimated_Q = kf.transition_covariance
            estimated_R = kf.observation_covariance

            smoothed_state_means, smoothed_state_covs = kf.smooth(data_window)
            smoothed_feature.extend(smoothed_state_means.flatten())
        smoothed_feature = np.array(smoothed_feature)
        return smoothed_feature

    @staticmethod
    def lsd_UnscentedKalmanFilter(data, window_size, observation_functions_type=None):
        """
        Apply an Unscented Kalman Filter for smoothing data.

        This function uses an Unscented Kalman Filter to smooth the input data within a specified window size.
        It supports optional sigmoid observation functions.

        Args:
            data (array-like): The input data to be smoothed.
            window_size (int): The size of the processing window.
            observation_functions_type (str, optional): The type of observation function. If 'sigmoid', applies a
                                                        sigmoid-based observation function. Defaults to None.

        Returns:
            ndarray: The smoothed data.
        """
        from pykalman import UnscentedKalmanFilter
        window_num = data.shape[0] // window_size
        smoothed_feature = []
        if observation_functions_type == "sigmoid":
            def measurement_function(x, w):
                return np.arctanh(x / (5 * 10 ** 7)) * 10 ** 7 + w

            def measurement_function_oly_x(x):
                return np.arctanh(x / (5 * 10 ** 7)) * 10 ** 7
        else:
            measurement_function = None
        for i_window in range(window_num + 1):
            begin_idx = window_size * i_window
            end_idx = window_size * (i_window + 1)
            if begin_idx >= data.shape[0]:
                continue
            if end_idx > data.shape[0]:
                end_idx = data.shape[0]
            data_window = data[begin_idx:end_idx]

            transition_covariance = 0.1
            observation_covariance = 0.001
            if observation_functions_type == "sigmoid":
                initial_state = [measurement_function_oly_x(x) for x in data_window]
                initial_state_mean = np.mean(initial_state)
            else:
                initial_state_mean = np.mean(data_window)
            initial_state_covariance = 1
            kf = UnscentedKalmanFilter(

                observation_functions=measurement_function,
                transition_covariance=transition_covariance,
                observation_covariance=observation_covariance,
                initial_state_mean=initial_state_mean,
                initial_state_covariance=initial_state_covariance
            )

            smoothed_state_means, smoothed_state_covs = kf.smooth(data_window)
            smoothed_feature.extend(smoothed_state_means.flatten())
        smoothed_feature = np.array(smoothed_feature)
        return smoothed_feature

    def feature_smooth(self, data, smooth_type="lds", window_size=10):
        """
        Smooth features using specified smoothing techniques.

        This function smooths the features using various smoothing techniques, including moving average filter,
        linear dynamical system (LDS) approach, and Unscented Kalman Filter. The method and window size for
        smoothing can be specified.

        Args:
            data (ndarray): The input data with shape (n_epoch, n_channel, n_feature).
            smooth_type (str, optional): The type of smoothing technique to use. Options are:
                - "mv_av_filter": Moving average filter
                - "lds": Linear dynamic system (LDS) approach
                - "UnscentedKalmanFilter": Unscented Kalman Filter
                - "UnscentedKalmanFilter_sigmoid": Unscented Kalman Filter with sigmoid function
                Defaults to "lds".
            window_size (int, optional): The size of the processing window. Defaults to 10.

        Returns:
            ndarray: The smoothed data.

        References:
            Duan R N, Zhu J Y, Lu B L. Differential entropy feature for EEG-based emotion classification[C]//2013 6th International IEEE/EMBS Conference on Neural Engineering (NER). IEEE, 2013: 81-84.
            Shi L C, Lu B L. Off-line and on-line vigilance estimation based on linear dynamical system and manifold learning[C]//2010 Annual International Conference of the IEEE Engineering in Medicine and Biology. IEEE, 2010: 6587-6590.
            Zheng W L, Zhu J Y, Lu B L. Identifying stable patterns over time for emotion recognition from EEG[J]. IEEE Transactions on Affective Computing, 2017, 10(3): 417-429.
        """
        n_eopoch, n_channel, n_feature = data.shape
        print(n_eopoch, n_channel, n_feature)
        smoothed_data = np.zeros((n_eopoch, n_channel, n_feature))
        for i_feature in range(n_feature):
            for i_channel in range(n_channel):
                print(data[:, i_channel, i_feature].shape)
                if smooth_type == "mv_av_filter":
                    smoothed_data[:, i_channel, i_feature] = self.moving_average_filter(data[:, i_channel, i_feature],
                                                                                        window_size)
                if smooth_type == "lds":
                    smoothed_data[:, i_channel, i_feature] = self.lsd_KalmanFilter(data[:, i_channel, i_feature],
                                                                                   window_size)
                if smooth_type == "UnscentedKalmanFilter":
                    smoothed_data[:, i_channel, i_feature] = self.lsd_UnscentedKalmanFilter(
                        data[:, i_channel, i_feature],
                        window_size)
                if smooth_type == "UnscentedKalmanFilter_sigmoid":
                    smoothed_data[:, i_channel, i_feature] = self.lsd_UnscentedKalmanFilter(
                        data[:, i_channel, i_feature], window_size, "sigmoid")
        return smoothed_data







#!/usr/bin/env python

from __future__ import division

from .cum3est import *
from .cum4est import *


def cumest(y, norder=2, maxlag=0, nsamp=None, overlap=0, flag='biased', k1=0, k2=0):
    """
  Second-, third- or fourth-order cumulants.
  Parameters:
           y - time-series  - should be a vector
      norder - cumulant order: 2, 3 or 4 [default = 2]
      maxlag - maximum cumulant lag to compute [default = 0]
    samp_seg - samples per segment  [default = data_length]
     overlap - percentage overlap of segments [default = 0]
               overlap is clipped to the allowed range of [0,99].
       flag  - 'biased' or 'unbiased'  [default = 'biased']
      k1,k2  - specify the slice of 3rd or 4th order cumulants

  Output:
      y_cum  - C2(m) or C3(m,k1) or C4(m,k1,k2),  -maxlag <= m <= maxlag
               depending upon the cumulant order selected
  """

    (ksamp, nrecs) = y.shape
    if ksamp == 1:
        ksamp = nrecs
        nrecs = 1

    if norder < 2 or norder > 4:
        raise ValueError('cumulant order must be 2, 3 or 4')

    if maxlag < 0:
        raise ValueError('"maxlag" must be non-negative')

    if nrecs > 1: nsamp = ksamp
    if nsamp <= 0 or nsamp > ksamp: nsamp = ksamp

    if nrecs > 1: overlap = 0
    overlap = max(0, min(overlap, 99))

    # estimate the cumulants
    if norder == 2:
        y_cum = cum2est(y, maxlag, nsamp, overlap, flag)
    elif norder == 3:
        y_cum = cum3est(y, maxlag, nsamp, overlap, flag, k1)
    elif norder == 4:
        y_cum = cum3est(y, maxlag, nsamp, overlap, flag, k1, k2)

    return y_cum


def test():
    y = sio.loadmat(here(__file__) + '/demo/ma1.mat')['y']

    # The right results are:
    #           "biased": [-0.12250513  0.35963613  1.00586945  0.35963613 -0.12250513]
    #           "unbiaed": [-0.12444965  0.36246791  1.00586945  0.36246791 -0.12444965]
    print(cum2est(y, 2, 128, 0, 'unbiased'))
    print(cum2est(y, 2, 128, 0, 'biased'))

    # For the 3rd cumulant:
    #           "biased": [-0.18203039  0.07751503  0.67113035  0.729953    0.07751503]
    #           "unbiased": [-0.18639911  0.07874543  0.67641484  0.74153955  0.07937539]
    print(cum3est(y, 2, 128, 0, 'biased', 1))
    print(cum3est(y, 2, 128, 0, 'unbiased', 1))

    # For testing the 4th-order cumulant
    # "biased": [-0.03642083  0.4755026   0.6352588   1.38975232  0.83791117  0.41641134 -0.97386322]
    # "unbiased": [-0.04011388  0.48736793  0.64948927  1.40734633  0.8445089   0.42303979 -0.99724968]
    print(cum4est(y, 3, 128, 0, 'biased', 1, 1))
    print(cum4est(y, 3, 128, 0, 'unbiased', 1, 1))


if __name__ == '__main__':
    test()
